import attr

from ..factory import target_factory
from ..step import step
from .common import Driver
from ..util.managedfile import ManagedFile
from ..util.helper import processwrapper


@target_factory.reg_driver
@attr.s(eq=False)
class DFUDriver(Driver):
    bindings = {
        "dfu": {"DFUDevice", "NetworkDFUDevice"},
    }

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        # FIXME make sure we always have an environment or config
        if self.target.env:
            self.tool = self.target.env.config.get_tool("dfu-util")
        else:
            self.tool = "dfu-util"

    def _get_dfu_prefix(self):
        return self.dfu.command_prefix + [
            self.tool,
            "-p",
            self.dfu.path,
        ]

    def on_activate(self):
        pass

    def on_deactivate(self):
        pass

    @Driver.check_active
    @step(args=["altsetting", "filename"])
    def download(self, altsetting, filename):
        mf = ManagedFile(filename, self.dfu)
        mf.sync_to_resource()

        processwrapper.check_output(
            self._get_dfu_prefix() + ["--alt", str(altsetting), "--download", mf.get_remote_path()],
            print_on_silent_log=True,
        )

    @step()
    def detach(self, altsetting):
        processwrapper.check_output(self._get_dfu_prefix() + ["--alt", str(altsetting), "--detach"])

    @step()
    def list(self):
        processwrapper.check_output(self._get_dfu_prefix() + ["--list"], print_on_silent_log=True)



import logging

import attr

from .step import steps, StepEvent
from .util import re_vt100

DEFAULT_FORMAT = "%(levelname)-7.7s %(name)15.15s: %(message)s"

def basicConfig(**kwargs):
    kwargs.setdefault("format", DEFAULT_FORMAT)
    indent = kwargs.get("indent", True)
    logging.basicConfig(**kwargs)
    root = logging.getLogger()

    parent = root.handlers[0].formatter
    root.handlers[0].setFormatter(StepFormatter(indent=indent, parent=parent))


logging.CONSOLE = logging.INFO - 5
assert(logging.CONSOLE > logging.DEBUG)
logging.addLevelName(logging.CONSOLE, "CONSOLE")

# Use composition instead of inheritance
class StepFormatter:
    def __init__(self, *args, indent=True, color=None, parent=None, **kwargs):
        self.formatter = parent or logging.Formatter(*args, **kwargs)
        self.indent = indent
        self.indent_level = 0
        self.bufs = dict()
        self.color = color

    def format(self, record):
        old_msg = record.msg
        try:
            if self.indent:
                if hasattr(record, "indent_level"):
                    self.indent_level = record.indent_level

                record.msg = (" " * self.indent_level) + str(record.msg)

                self.indent_level = getattr(
                    record, "next_indent_level", self.indent_level
                )

            if hasattr(record, "step"):
                record.pathname, record.filename, record.lineno = record.step.sourceinfo

            return self.formatter.format(record)
        finally:
            record.msg = old_msg


@attr.s
class SerialLoggingReporter:
    bufs = attr.ib(
        default=attr.Factory(dict), validator=attr.validators.instance_of(dict)
    )
    loggers = attr.ib(
        default=attr.Factory(dict), validator=attr.validators.instance_of(dict)
    )
    lastevent = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(StepEvent)),
    )

    def __attrs_post_init__(self):
        steps.subscribe(self.notify)

    def vt100_replace_cr_nl(self, buf):
        string = re_vt100.sub("", buf.decode("utf-8", errors="replace"))
        string = string.replace("\r", "␍")
        string = string.replace("\n", "␤")
        string = string.replace("\b", "␈")
        string = string.replace("\a", "␇")
        string = string.replace("\v", "␋")
        string = string.replace("\f", "␌")
        return string

    def _create_message(self, event, data):
        return "{source} {dirind} {data}␍␤".format(
            source=event.step.source,
            dirind="<" if event.step.title == "read" else ">",
            data=data,
        )

    def notify(self, event):
        step = event.step
        state = event.data.get("state")
        extra = {
            "step": step,
        }
        if step.tag == "console":
            self.loggers[step.source] = logging.getLogger(
                f"SerialLogger.{step.source.target.name}.{step.source.__class__.__name__}"
            )
            logger = self.loggers[step.source]

            if state == "stop" and step.title == "read" and step.result:
                if step.source not in self.bufs.keys():
                    self.bufs[step.source] = b""
                self.bufs[step.source] += step.result
                *parts, self.bufs[step.source] = self.bufs[step.source].split(b"\r\n")

                self.lastevent = event

                for part in parts:
                    data = self.vt100_replace_cr_nl(part)
                    logger.log(logging.CONSOLE, self._create_message(event, data), extra=extra)

            elif state == "start" and step.args and "data" in step.args:
                data = self.vt100_replace_cr_nl(step.args["data"])
                logger.log(logging.CONSOLE, self._create_message(event, data), extra=extra)

    def flush(self):
        if self.lastevent is None:
            return

        extra = {
            "step": self.lastevent.step,
        }
        for source, logger in self.loggers.items():
            data = self.vt100_replace_cr_nl(self.bufs[source])
            if data:
                logger.log(logging.CONSOLE, self._create_message(self.lastevent, data), extra=extra)
            self.bufs[source] = b""


class StepLogger:
    _started = False
    _logger = None
    _serial_logger = None
    _length_limit = 100

    def __attrs_post_init__(self):
        from warnings import warn

        warn(
            "StepLogger should not be instantiated, use StepLogger.start()/.stop() instead.",
            DeprecationWarning,
            stacklevel=2,
        )

    @classmethod
    def start(cls, length_limit=None):
        """starts the StepLogger"""
        assert not cls._started
        if cls._logger is None:
            cls._logger = logging.getLogger("StepLogger")
        steps.subscribe(cls.notify)
        cls._serial_logger = SerialLoggingReporter()
        cls._started = True
        if length_limit is not None:
            cls._length_limit = length_limit

    @classmethod
    def stop(cls):
        """stops the StepLogger"""
        assert cls._started
        steps.unsubscribe(cls.notify)
        cls._started = False

    @classmethod
    def get_prefix(cls, event):
        if event.step.exception:
            return "⚠"
        elif event.data.get("state") == "start":
            return "→"
        elif event.data.get("state") == "stop":
            return "←"
        elif event.data.get("skip", None):
            return "↓"
        else:
            return ""

    @staticmethod
    def format_arguments(args):
        if args is None:
            return ""
        if isinstance(args, dict):
            collected_args = []
            for k, v in args.items():
                collected_args.append(f"{k}={repr(v)}")

            return " ".join(collected_args)
        else:
            return "{}".format(args)

    @staticmethod
    def format_duration(duration):
        if duration < 0.001:
            return ""

        return "[{:.3f}s]".format(duration)

    @classmethod
    def format_result(cls, result):
        if result is None:
            return ""

        if len(str(result)) < cls._length_limit or cls._length_limit is None:
            return "result={} ".format(result)

        return f"result={repr(result):.{cls._length_limit}}… "

    @classmethod
    def __get_message(cls, event):
        m = "{prefix} {source}.{title}({args}) {result}{duration}".format(
            prefix=cls.get_prefix(event),
            source=event.step.source.__class__.__name__,
            title=event.step.title,
            args=cls.format_arguments(event.data.get("args", {})),
            result=cls.format_result(event.data.get("result", None)),
            duration=cls.format_duration(event.data.get("duration", 0.0)),
        )

        if event.step.exception:
            m += f" exception={event.step.exception}"

        reason = event.data.get("skip", None)
        if reason:
            m += f"skipped={reason}"

        return m

    @classmethod
    def get_next_indent(cls, event):
        if event.step.exception:
            return event.step.level

        if event.data.get("state") == "start":
            return event.step.level + 1

        return event.step.level

    @classmethod
    def notify(cls, event):
        if event.step.tag == "console":
            return
        if cls._serial_logger:
            cls._serial_logger.flush()

        message = cls.__get_message(event)

        level = logging.INFO
        extra = {
            "indent_level": event.step.level,
            "step": event.step,
            "next_indent_level": cls.get_next_indent(event),
        }
        cls._logger.log(level, message, extra=extra)



from labgrid.resource import ManagedResource, Resource


def test_create_resource(target):
    resource = Resource(target, "resource")


def test_create_managed_resource(target):
    resource_1 = ManagedResource(target, "managedr1")
    resource_2 = ManagedResource(target, "managedr2")
    assert resource_1.manager is resource_2.manager


def test_hash_is(target):
    resource1 = Resource(target, "resource1")
    resource2 = Resource(target, "resource2")
    assert resource1 is not resource2


def test_hash_set(target):
    resource1 = Resource(target, "resource")
    k = set()
    k.add(resource1)
    assert resource1 in k



from importlib import import_module
import attr

from ..factory import target_factory
from .common import Driver


@target_factory.reg_driver
@attr.s(eq=False)
class XenaDriver(Driver):
    """
    Xena Driver
    """

    bindings = {"xena_manager": "XenaManager"}

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self._xena_app = import_module("xenavalkyrie.xena_app")
        self._tgn_utils = import_module("trafficgenerator.tgn_utils")
        self._xm = None

    def on_activate(self):
        self._xm = self._xena_app.init_xena(self._tgn_utils.ApiType.socket, self.logger, "labgrid")
        self._xm.session.add_chassis(self.xena_manager.hostname)

    def on_deactivate(self):
        if self._xm:
            self._xm.session.disconnect()
            self._xm = None

    @Driver.check_active
    def get_session(self):
        return self._xm.session



#!/usr/bin/env python3
# Generates an Ethernet frame via scapy using pcap, copies pcap to DUT, replays pcap on interface,
# records frame locally (or on exporter, adjust env.yaml accordingly), and compares both.

import logging
import os
from tempfile import NamedTemporaryFile, TemporaryDirectory

from scapy.all import Ether, Raw, rdpcap, wrpcap, conf

from labgrid import Environment
from labgrid.logging import basicConfig, StepLogger


def generate_frame():
    frame = Ether(dst="11:22:33:44:55:66", src="66:55:44:33:22:11", type=0x9000)
    padding = "\x00" * (conf.min_pkt_size - len(frame))
    frame /= Raw(load=padding)
    return frame


basicConfig(level=logging.INFO)
StepLogger.start()
env = Environment("env.yaml")
target = env.get_target()

netdrv = target.get_driver("RawNetworkInterfaceDriver")
ssh = target.get_driver("SSHDriver")

# get DUT interface
exporter_iface = netdrv.iface.ifname
dut_iface = env.config.get_target_option(target.name, "local_iface_to_dut_iface")[exporter_iface]

# generate test frame
generated_frame = generate_frame()

# write pcap, copy to DUT
remote_pcap = "/tmp/pcap"
with NamedTemporaryFile() as pcap:
    wrpcap(pcap.name, generated_frame)
    ssh.put(pcap.name, remote_pcap)

# copy recorded pcap from DUT, compare with generated frame
with TemporaryDirectory() as tempdir:
    # start record on exporter
    tempf = os.path.join(tempdir, "record.pcap")
    with netdrv.record(tempf, count=1) as record:
        # replay pcap on DUT
        ssh.run_check(f"ip link set {dut_iface} up")
        ssh.run_check(f"tcpreplay -i {dut_iface} {remote_pcap}")

    remote_frame = rdpcap(tempf)
    assert remote_frame[0] == generated_frame[0]

print("statistics", netdrv.get_statistics())
print("address", netdrv.get_address())



import os.path
import attr

from ..resource import NetworkDediprogFlasher
from ..factory import target_factory
from ..step import step
from .common import Driver, check_file
from ..util.managedfile import ManagedFile
from ..util.helper import processwrapper


@target_factory.reg_driver
@attr.s(eq=False)
class DediprogFlashDriver(Driver):
    """The DediprogFlashDriver uses the dediprog utility to write an image
    to a raw rom. The driver is a pure wrapper of the dpcmd utility"""

    bindings = {
        "flasher": {"DediprogFlasher", NetworkDediprogFlasher},
    }

    image = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(str)), default=None)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        if self.target.env:
            self.tool = self.target.env.config.get_tool("dpcmd")
        else:
            self.tool = "dpcmd"

    def _get_dediprog_prefix(self):
        return self.flasher.command_prefix + [self.tool]

    def on_activate(self):
        pass

    def on_deactivate(self):
        pass

    def map_vcc(self):
        vcc_map = {"3.5V": "0", "2.5V": "1", "1.8V": "2"}
        return vcc_map[self.flasher.vcc]

    @Driver.check_active
    @step(title="call", args=["args"])
    def __call__(self, *args):
        vcc = self.map_vcc()
        arg_list = list(args)
        arg_list.append("--vcc")
        arg_list.append(vcc)
        arg_list.append("--silent")
        processwrapper.check_output(self._get_dediprog_prefix() + arg_list)

    @Driver.check_active
    @step(args=["filename"])
    def flash(self, filename=None):
        if filename is None and self.image is not None:
            filename = self.target.env.config.get_image_path(self.image)
        filename = os.path.abspath(filename)
        check_file(filename)
        mf = ManagedFile(filename, self.flasher)
        mf.sync_to_resource()

        self("--auto", mf.get_remote_path(), "--verify", "-x", "ff")

    @Driver.check_active
    def erase(self):
        self("--erase")



import os.path
import re
import subprocess
import socket
import atexit
import warnings

from shutil import which

import attr
import pytest
import logging

from labgrid.util import diff_dict, flat_dict, filter_dict
from labgrid.util.helper import get_free_port
from labgrid.util.ssh import ForwardError, SSHConnection, sshmanager
from labgrid.util.proxy import proxymanager
from labgrid.util.managedfile import ManagedFile, ManagedFileError
from labgrid.driver.exception import ExecutionError
from labgrid.resource.serialport import NetworkSerialPort
from labgrid.resource.common import Resource, NetworkResource
from labgrid.util import diff_dict, flat_dict, filter_dict, find_dict

@pytest.fixture
def connection_localhost():
    con = SSHConnection("localhost")
    con.connect()
    yield con
    con.disconnect()

@pytest.fixture
def connection_localhost_no_cleanup():
    con = SSHConnection("localhost")
    con.connect()
    return con

@pytest.fixture
def sshmanager_fix():
    yield sshmanager
    sshmanager.close_all()

def test_diff_dict():
    dict_a = {"a": 1,
              "b": 2}
    dict_b = {"a": 1,
              "b": 3}
    gen = diff_dict(dict_a, dict_b)
    for res in gen:
        assert res[0] == 'b'
        assert res[1] == 2
        assert res[2] == 3

def test_flat_dict():
    dict_a = {"a":
              {"b": 3},
              "b": 2}
    res = flat_dict(dict_a)
    assert res == {"a.b": 3, "b": 2}

def test_filter_dict():
    @attr.s
    class A:
        foo = attr.ib()

    d_orig = {'foo': 1, 'bar': 2, 'baz': 3}

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        d_filtered = filter_dict(d_orig, A)
    assert d_filtered is not d_orig
    assert d_filtered == {'foo': 1}

    with pytest.warns(UserWarning) as record:
        d_filtered = filter_dict(d_orig, A, warn=True)
    assert len(record) == 2
    assert str(record[0].message) == "unsupported attribute 'bar' with value '2' for class 'A'"
    assert str(record[1].message) == "unsupported attribute 'baz' with value '3' for class 'A'"
    assert d_filtered is not d_orig
    assert d_filtered == {'foo': 1}

def test_sshmanager_get():
    assert sshmanager != None

def test_sshconnection_get():
    from labgrid.util.ssh import SSHConnection
    SSHConnection("localhost")

@pytest.mark.skipif(not which("ssh"), reason="ssh not available")
def test_sshconnection_inactive_raise():
    from labgrid.util.ssh import SSHConnection
    con = SSHConnection("localhost")
    with pytest.raises(ExecutionError):
        con.run_check("echo Hallo")

@pytest.mark.localsshmanager
def test_sshconnection_connect(connection_localhost):
    assert connection_localhost.isconnected()
    assert os.path.exists(connection_localhost._socket)

@pytest.mark.localsshmanager
def test_sshconnection_run(connection_localhost):
    stdout, stderr, exitcode = connection_localhost.run("echo stderr >&2; echo stdout1; echo stdout2")
    assert exitcode == 0
    assert stderr == ["stderr"]
    assert stdout == ["stdout1", "stdout2"]

@pytest.mark.localsshmanager
def test_sshconnection_run_log(connection_localhost, caplog):
    caplog.set_level(logging.INFO)
    stdout, stderr, exitcode = connection_localhost.run("echo stderr >&2; echo stdout1; echo stdout2",
            stdout_loglevel=logging.INFO, stderr_loglevel=logging.WARNING)
    assert exitcode == 0
    assert stderr == ["stderr"]
    assert stdout == ["stdout1", "stdout2"]
    assert sorted([(rec[1], rec[2]) for rec in caplog.record_tuples]) == [
        (logging.INFO, 'stdout1'),
        (logging.INFO, 'stdout2'),
        (logging.WARNING, 'stderr'),
   ]

@pytest.mark.localsshmanager
def test_sshconnection_run_merged_stderr(connection_localhost):
    stdout, stderr, exitcode = connection_localhost.run(
            "echo stderr >&2; echo stdout", stderr_merge=True)
    assert exitcode == 0
    assert sorted(stdout) == ["stderr", "stdout"]
    assert stderr == []

@pytest.mark.localsshmanager
def test_sshconnection_run_fail(connection_localhost):
    stdout, stderr, exitcode = connection_localhost.run("false")
    assert exitcode != 0


@pytest.mark.localsshmanager
def test_sshconnection_port_forward_add_remove(connection_localhost):
    port = get_free_port()
    test_string = "Hello World"

    local_port = connection_localhost.add_port_forward('localhost', port)
    server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    server_socket.bind(("127.0.0.1", port))
    server_socket.listen(1)
    send_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    send_socket.connect(("127.0.0.1", local_port))
    client_socket, address = server_socket.accept()
    send_socket.send(test_string.encode('utf-8'))

    assert client_socket.recv(16).decode("utf-8") == test_string
    connection_localhost.remove_port_forward('localhost', port)

@pytest.mark.localsshmanager
def test_sshconnection_port_forward_remove_raise(connection_localhost):
    port = get_free_port()

    with pytest.raises(ForwardError):
        connection_localhost.remove_port_forward('localhost', port)

@pytest.mark.localsshmanager
def test_sshconnection_port_forward_add_duplicate(connection_localhost):
    port = get_free_port()

    first_port = connection_localhost.add_port_forward('localhost', port)
    second_port = connection_localhost.add_port_forward('localhost', port)
    assert first_port == second_port


@pytest.mark.localsshmanager
def test_sshconnection_port_remote_forward_add_remove(connection_localhost):
    rport = get_free_port()
    lport = get_free_port()
    test_string = "Hello World"

    connection_localhost.add_remote_port_forward(rport, lport)
    server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    server_socket.bind(("127.0.0.1", lport))
    server_socket.listen(1)
    send_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    send_socket.connect(("127.0.0.1", rport))
    client_socket, address = server_socket.accept()
    send_socket.send(test_string.encode('utf-8'))

    assert client_socket.recv(16).decode("utf-8") == test_string
    connection_localhost.remove_remote_port_forward(rport, lport)
    assert connection_localhost._r_forwards == set()


@pytest.mark.localsshmanager
def test_sshconnection_put_file(connection_localhost, tmpdir):
    p = tmpdir.join("config.yaml")
    p.write(
        """Teststring"""
    )
    connection_localhost.put_file(p, '/tmp/test')

    assert os.path.isfile('/tmp/test')
    assert open('/tmp/test', 'r').readlines() == [ "Teststring" ]

@pytest.mark.localsshmanager
def test_sshconnection_get_file(connection_localhost, tmpdir):

    p = tmpdir.join("test")
    connection_localhost.get_file('/tmp/test', p)

@pytest.mark.localsshmanager
def test_sshconnection_cleanup(connection_localhost_no_cleanup):

    connection_localhost_no_cleanup.cleanup()

    # do not call cleanup() again on termination
    atexit.unregister(connection_localhost_no_cleanup.cleanup)

@pytest.mark.localsshmanager
def test_sshmanager_open(sshmanager_fix):
    con = sshmanager_fix.open("localhost")

    assert isinstance(con, SSHConnection)

@pytest.mark.localsshmanager
def test_sshmanager_add_forward(sshmanager_fix):
    port = sshmanager_fix.request_forward("localhost", 'localhost', 3000)

    assert port < 65536

@pytest.mark.localsshmanager
def test_sshmanager_remove_forward(sshmanager_fix):
    port = sshmanager_fix.request_forward("localhost", 'localhost', 3000)
    sshmanager_fix.remove_forward('localhost', 'localhost', 3000)

    assert 3000 not in sshmanager_fix.get('localhost')._l_forwards

@pytest.mark.localsshmanager
def test_sshmanager_close(sshmanager_fix):
    con = sshmanager_fix.open("localhost")

    assert isinstance(con, SSHConnection)
    sshmanager_fix.close("localhost")

@pytest.mark.localsshmanager
def test_sshmanager_remove_raise(sshmanager_fix):
    con = sshmanager_fix.open("localhost")
    con.connect()
    with pytest.raises(ExecutionError):
        sshmanager_fix.remove_connection(con)

@pytest.mark.skipif(not which("ssh"), reason="ssh not available")
def test_sshmanager_add_duplicate(sshmanager_fix):
    host = 'localhost'
    con = SSHConnection(host)
    sshmanager_fix.add_connection(con)
    con_there = sshmanager_fix._connections[host]
    sshmanager_fix.add_connection(con)
    con_now = sshmanager_fix._connections[host]

    assert con_now == con_there

@pytest.mark.skipif(not which("ssh"), reason="ssh not available")
def test_sshmanager_add_new(sshmanager_fix):
    host = 'other_host'
    con = SSHConnection(host)
    sshmanager_fix.add_connection(con)
    con_now = sshmanager_fix._connections[host]

    assert con_now == con

@pytest.mark.skipif(not which("ssh"), reason="ssh not available")
def test_sshmanager_invalid_host_raise():
    con = SSHConnection("nosuchhost.notavailable")
    with pytest.raises(ExecutionError):
        con.connect()

def test_proxymanager_no_proxy(target):
    host = 'localhost'
    port = 5000
    nr = NetworkSerialPort(target, host=host, port=port, name=None)
    nr.proxy = ''
    nr.proxy_required = False

    assert (host, port) == proxymanager.get_host_and_port(nr)

@pytest.mark.localsshmanager
def test_proxymanager_remote_forced_proxy(target):
    sshmanager.close_all()
    host = 'localhost'
    port = 5000
    nr = NetworkSerialPort(target, host=host, port=port, name=None)
    extras = { 'proxy': 'localhost', 'proxy_required': True }
    nr.extra = extras
    nhost, nport = proxymanager.get_host_and_port(nr)

    assert (host, port) != (nhost, nport)
    sshmanager.close_all()

@pytest.mark.localsshmanager
def test_proxymanager_local_forced_proxy(target):
    host = 'localhost'
    port = 5000
    nr = NetworkSerialPort(target, host=host, port=port, name=None)
    extras = { 'proxy': 'localhost', 'proxy_required': True }
    nr.extra = extras
    nhost, nport = proxymanager.get_host_and_port(nr)

    assert (host, port) != (nhost, nport)

@pytest.mark.localsshmanager
def test_remote_managedfile(target, tmpdir):
    import hashlib
    import getpass

    res = NetworkResource(target, "test", "localhost")
    t = tmpdir.join("test")
    t.write(
"""
Test
"""
    )
    hash = hashlib.sha256(t.read().encode("utf-8")).hexdigest()
    mf = ManagedFile(t, res, detect_nfs=False)
    mf.sync_to_resource()

    assert os.path.isfile(f"/var/cache/labgrid/{getpass.getuser()}/{hash}/test")
    assert hash == mf.get_hash()
    assert f"/var/cache/labgrid/{getpass.getuser()}/{hash}/test" == mf.get_remote_path()

@pytest.mark.localsshmanager
def test_remote_managedfile_on_nfs(target, tmpdir):
    res = NetworkResource(target, "test", "localhost")
    t = tmpdir.join("test")
    t.write(
"""
Test
"""
    )
    mf = ManagedFile(t, res, detect_nfs=True)
    mf.sync_to_resource()

    assert str(t) == mf.get_remote_path()

def test_local_managedfile(target, tmpdir):
    import hashlib

    res = Resource(target, "test")
    t = tmpdir.join("test")
    t.write(
"""
Test
"""
    )
    hash = hashlib.sha256(t.read().encode("utf-8")).hexdigest()
    mf = ManagedFile(t, res, detect_nfs=False)
    mf.sync_to_resource()

    assert hash == mf.get_hash()
    assert str(t) == mf.get_remote_path()

def test_network_managedfile_no_sync(target, tmpdir):
    res = NetworkResource(target, "test", "localhost")
    t = tmpdir.join("test")
    t.write(
"""
Test
"""
    )
    mf = ManagedFile(t, res, detect_nfs=False)

    expected = re.escape("sync_to_resource() needs to be called before the remote-path can be retrieved")
    with pytest.raises(ManagedFileError, match=expected) as e:
        mf.get_remote_path()

def test_local_managedfile_symlink(target, tmpdir):
    res = Resource(target, "test")
    t = tmpdir.join("test")
    t.write(
"""
Test
"""
    )
    mf = ManagedFile(t, res, detect_nfs=False)
    mf.sync_to_resource(symlink=tmpdir.join("link"))

    assert str(t) == mf.get_remote_path()
    assert os.path.islink(tmpdir.join("link"))

@pytest.mark.localsshmanager
def test_remote_managedfile_symlink(target, tmpdir):
    res = NetworkResource(target, "test", "localhost")
    t = tmpdir.join("test")
    t.write(
"""
Test
"""
    )
    mf = ManagedFile(t, res, detect_nfs=False)
    mf.sync_to_resource(symlink=tmpdir.join("link"))

    assert os.path.islink(tmpdir.join("link"))

def test_find_dict():
    dict_a = {"a": {"a.a": {"a.a.a": "a.a.a_val"}}, "b": "b_val"}
    assert find_dict(dict_a, "b") == "b_val"
    assert find_dict(dict_a, "a") == {"a.a": {"a.a.a": "a.a.a_val"}}
    assert find_dict(dict_a, "a.a") == {"a.a.a": "a.a.a_val"}
    assert find_dict(dict_a, "a.a.a") == "a.a.a_val"
    assert find_dict(dict_a, "x") == None



import abc


class MMIOProtocol(abc.ABC):
    @abc.abstractmethod
    def read(self, address: int, size: int, count: int) -> bytes:
        raise NotImplementedError

    @abc.abstractmethod
    def write(self, address: int, size: int, data: bytes) -> None:
        raise NotImplementedError



import attr

from ..factory import target_factory
from ..step import step
from .common import Driver
from ..exceptions import InvalidConfigError
from ..util.managedfile import ManagedFile
from ..util.helper import processwrapper
from ..resource.udev import AndroidUSBFastboot
from ..resource.fastboot import AndroidNetFastboot
from ..resource.remote import RemoteAndroidUSBFastboot, RemoteAndroidNetFastboot


@target_factory.reg_driver
@attr.s(eq=False)
class AndroidFastbootDriver(Driver):
    bindings = {
        "fastboot": {AndroidUSBFastboot, RemoteAndroidUSBFastboot,
                     AndroidNetFastboot, RemoteAndroidNetFastboot},
    }

    boot_image = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str))
    )
    flash_images = attr.ib(
        default={},
        validator=attr.validators.optional(attr.validators.instance_of(dict))
    )
    sparse_size = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str))
    )

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        # FIXME make sure we always have an environment or config
        if self.target.env:
            self.tool = self.target.env.config.get_tool('fastboot')
        else:
            self.tool = 'fastboot'

    def _get_fastboot_prefix(self):
        if isinstance(self.fastboot, (AndroidUSBFastboot, RemoteAndroidUSBFastboot)):
            option = f"usb:{self.fastboot.path}"
        else:
            option = f"{self.fastboot.protocol}:{self.fastboot.address}:{self.fastboot.port}"

        prefix = self.fastboot.command_prefix + [ self.tool, "-s", option ]

        if self.sparse_size is not None:
            prefix += ["-S", self.sparse_size]

        return prefix

    @staticmethod
    def _filter_fastboot_output(output, prefix='(bootloader) '):
        """
        Splits output by '\n' and returns only elements starting with prefix. The prefix is
        removed.
        """
        return [line[len(prefix):] for line in output.decode().split('\n') if line.startswith(prefix)]

    def on_activate(self):
        pass

    def on_deactivate(self):
        pass

    @Driver.check_active
    @step(title='call', args=['args'])
    def __call__(self, *args):
        processwrapper.check_output(
            self._get_fastboot_prefix() + list(args),
            print_on_silent_log=True
        )

    @Driver.check_active
    @step(args=['filename'])
    def boot(self, filename=None):
        if filename is None:
            if self.boot_image is None:
                raise InvalidConfigError("No boot_image set")

            filename = self.target.env.config.get_image_path(self.boot_image)

        mf = ManagedFile(filename, self.fastboot)
        mf.sync_to_resource()
        self("boot", mf.get_remote_path())

    @Driver.check_active
    @step(args=['partition', 'filename'])
    def flash(self, partition, filename=None):
        if filename is None:
            try:
                image_key = self.flash_images[partition]
            except KeyError:
                raise InvalidConfigError(f"Partition {partition} not in flash_images")

            filename = self.target.env.config.get_image_path(image_key)

        mf = ManagedFile(filename, self.fastboot)
        mf.sync_to_resource()
        self("flash", partition, mf.get_remote_path())

    @Driver.check_active
    @step()
    def flash_all(self):
        for partition in self.flash_images.keys():
            self.flash(partition)

    @Driver.check_active
    @step(args=['partition'])
    def erase(self, partition):
        self('erase', partition)

    @Driver.check_active
    @step(args=['cmd'])
    def run(self, cmd):
        self("oem", "exec", "--", cmd)

    @Driver.check_active
    @step(title='continue')
    def continue_boot(self):
        self("continue")

    @Driver.check_active
    @step(args=['var'])
    def getvar(self, var):
        """Return variable value via 'fastboot getvar <var>'."""
        if var == 'all':
            raise NotImplementedError('Retrieving a list of all variables is not supported yet')

        cmd = ['getvar', var]
        output = processwrapper.check_output(self._get_fastboot_prefix() + cmd)
        values = AndroidFastbootDriver._filter_fastboot_output(
            output, f'{var}: '
        )
        assert len(values) == 1, 'fastboot did not return exactly one line'
        return values[0]

    @Driver.check_active
    @step(args=['var'])
    def oem_getenv(self, var):
        """Return barebox environment variable value via 'fastboot oem getenv <var>'."""
        cmd = ['oem', 'getenv', var]
        output = processwrapper.check_output(self._get_fastboot_prefix() + cmd)
        values = AndroidFastbootDriver._filter_fastboot_output(output)
        assert len(values) == 1, 'fastboot did not return exactly one line'
        return values[0]



import subprocess
import re

import attr

from .common import Driver
from ..factory import target_factory
from ..step import step
from .exception import ExecutionError
from ..util.helper import processwrapper


@target_factory.reg_driver
@attr.s(eq=False)
class USBSDWire3Driver(Driver):
    """The USBSDWire3Driver uses the sdwire tool to control SDWire hardware

    Args:
        bindings (dict): driver to use with usbsdmux
    """

    bindings = {
        "mux": {"USBSDWire3Device", "NetworkUSBSDWire3Device"},
    }

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        if self.target.env:
            self.tool = self.target.env.config.get_tool("sdwire")
        else:
            self.tool = "sdwire"
        if self.mux.control_serial is None:
            raise ExecutionError("USBSDWire3Driver requires 'control_serial' to be set in the resource")
        self.control_serial = self.match_control_serial()

    @Driver.check_active
    @step(title="sdmux_set", args=["mode"])
    def set_mode(self, mode):
        if not mode.lower() in ["dut", "host"]:
            raise ExecutionError(f"Setting mode '{mode}' not supported by USBSDWire3Driver")
        cmd = self.mux.command_prefix + [
            self.tool,
            "switch",
            "-s",
            self.control_serial,
            "dut" if mode.lower() == "dut" else "ts",
        ]
        processwrapper.check_output(cmd)

    def match_control_serial(self):
        cmd = self.mux.command_prefix + [self.tool, "list"]
        proc = subprocess.run(cmd, stdout=subprocess.PIPE, check=True)
        output = proc.stdout.strip().decode()
        for line in output.splitlines():
            if self.mux.control_serial is not None and line.find(self.mux.control_serial) >= 0:
                return line.split()[0]
        raise ExecutionError(f"Could not find control serial {self.mux.control_serial} in sdwire list output")

    @Driver.check_active
    @step(title="sdmux_get")
    def get_mode(self):
        cmd = self.mux.command_prefix + [
            self.tool,
            "state",
            "-s",
            self.control_serial,
        ]
        result = processwrapper.check_output(cmd)
        for line in result.decode().splitlines():
            if re.match(self.control_serial, line):
                return line.split(" ", maxsplit=1)[1].strip()
        raise ExecutionError(f"Could not find control serial {self.mux.control_serial} in sdwire list output")



import os
import sys
from datetime import datetime

from .step import steps


class ConsoleLoggingReporter:
    """ConsoleLoggingReporter - Reporter that writes console log files

    Args:
        logpath (str): path to store the logfiles in
    """

    instance = None

    @classmethod
    def start(cls, path):
        """starts the ConsoleLoggingReporter"""
        assert cls.instance is None
        cls.instance = cls(path)

    @classmethod
    def stop(cls):
        """stops the ConsoleLoggingReporter"""
        assert cls.instance is not None
        cls.instance._stop()
        steps.unsubscribe(cls.instance.notify)
        cls.instance = None

    def __init__(self, logpath):
        self._logcache = {}
        self.logpath = logpath
        if not os.path.exists(self.logpath):
            os.makedirs(self.logpath)
        steps.subscribe(self.notify)

    def _stop(self):
        while self._logcache:
            _, log = self._logcache.popitem()
            # ignore cache entries for errors
            if log is None:
                continue
            log.close()

    def get_logfile(self, event):
        """Returns the correct file handle from cache or creates a new file handle"""
        source = event.step.source
        try:
            log = self._logcache[source]
        except KeyError:
            if source.name:
                name = f"console_{source.target.name}_{source.name}"
            else:
                name = f"console_{source.target.name}"
            name = os.path.join(self.logpath, name)
            try:
                log = self._logcache[source] = open(name, mode="ab", buffering=0)
            except OSError as e:
                print(f"failed to open log file {name}: {e}", file=sys.stderr)
                log = self._logcache[source] = None
            if not log:
                return None

            if source.name:
                log.write(f"Labgrid Console Logfile for {source.target.name} {source.name}\n".encode("utf-8"))
            else:
                log.write(f"Labgrid Console Logfile for {source.target.name}\n".encode("utf-8"))
            log.write(f"Logfile started at {datetime.now()}\n".encode("utf-8"))
            log.write("=== Log starts here ===\n".encode("utf-8"))

        return log

    def notify(self, event):
        """This is the callback function for steps"""
        step = event.step
        if step.tag == "console":
            if step.title == "read":
                if event.data.get("state") == "stop":
                    if step.result and step.source:
                        log = self.get_logfile(event)
                        if not log:
                            return
                        log.write(step.result)



import queue
import warnings
from collections import OrderedDict
from importlib import import_module

import attr

from ..factory import target_factory
from .common import ManagedResource, ResourceManager
from .base import SerialPort, NetworkInterface
from ..util import Timeout


@attr.s(eq=False)
class UdevManager(ResourceManager):
    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self.queue = queue.Queue()

        self._pyudev = import_module('pyudev')
        self._context = self._pyudev.Context()
        self._monitor = self._pyudev.Monitor.from_netlink(self._context)
        self._observer = self._pyudev.MonitorObserver(self._monitor,
                                                callback=self._insert_into_queue)
        self._observer.start()

    def on_resource_added(self, resource):
        devices = self._context.list_devices()
        devices.match_subsystem(resource.match['SUBSYSTEM'])
        for device in devices:
            if resource.try_match(device):
                self.logger.debug(" matched successfully against %s", resource.device)

    def _insert_into_queue(self, device):
        self.queue.put(device)

    def poll(self):
        timeout = Timeout(0.1)
        while not timeout.expired:
            try:
                device = self.queue.get(False)
            except queue.Empty:
                break
            self.logger.debug("%s: %s", device.action, device)
            for resource in self.resources:
                if resource.try_match(device):
                    self.logger.debug(" matched successfully")

@attr.s(eq=False)
class USBResource(ManagedResource):
    manager_cls = UdevManager

    match = attr.ib(factory=dict, validator=attr.validators.instance_of(dict), hash=False)
    device = attr.ib(default=None, hash=False)
    suggest = attr.ib(default=False, hash=False, repr=False)

    def __attrs_post_init__(self):
        self.timeout = 5.0
        self.match.setdefault('SUBSYSTEM', 'usb')
        super().__attrs_post_init__()

    def filter_match(self, device):  # pylint: disable=unused-argument
        return True

    def suggest_match(self, device):
        meta = OrderedDict()
        suggestions = []

        if self.device.device_node:
            meta['device node'] = self.device.device_node
        if list(self.device.tags):
            meta['udev tags'] = ', '.join(self.device.tags)
        if self.device.properties.get('ID_VENDOR'):
            meta['vendor'] = self.device.properties.get('ID_VENDOR')
        if self.device.properties.get('ID_VENDOR_FROM_DATABASE'):
            meta['vendor (DB)'] = self.device.properties.get('ID_VENDOR_FROM_DATABASE')
        if self.device.properties.get('ID_MODEL'):
            meta['model'] = self.device.properties.get('ID_MODEL')
        if self.device.properties.get('ID_MODEL_FROM_DATABASE'):
            meta['model (DB)'] = self.device.properties.get('ID_MODEL_FROM_DATABASE')
        if self.device.properties.get('ID_REVISION'):
            meta['revision'] = self.device.properties.get('ID_REVISION')

        path = self.device.properties.get('ID_PATH')
        if path:
            suggestions.append({'ID_PATH': path})
        elif self.match.get('@SUBSYSTEM', None) == 'usb':
            path = self._get_usb_device().properties.get('ID_PATH')
            if path:
                suggestions.append({'@ID_PATH': path})

        serial = self.device.properties.get('ID_SERIAL_SHORT')
        interface_num = self.device.properties.get('ID_USB_INTERFACE_NUM')
        if serial:
            if interface_num is not None:
                suggestions.append({'ID_SERIAL_SHORT': serial,
                                    'ID_USB_INTERFACE_NUM': interface_num})
            else:
                suggestions.append({'ID_SERIAL_SHORT': serial})
        elif self.match.get('@SUBSYSTEM', None) == 'usb':
            serial = self._get_usb_device().properties.get('ID_SERIAL_SHORT')
            interface_num = self._get_usb_device().properties.get('ID_USB_INTERFACE_NUM')
            if serial:
                if interface_num is not None:
                    suggestions.append({'@ID_SERIAL_SHORT': serial,
                                        '@ID_USB_INTERFACE_NUM': interface_num})
                else:
                    suggestions.append({'@ID_SERIAL_SHORT': serial})

        return meta, suggestions

    def try_match(self, device):
        if self.device is None:  # new device
            def match_single(dev, key, value):
                if dev.properties.get(key) == value:
                    return True
                if dev.attributes.get(key) and dev.attributes.asstring(key) == value:
                    return True
                if getattr(dev, key, None) == value:
                    return True
                return False

            def match_ancestors(key, value):
                for ancestor in device.ancestors:
                    if match_single(ancestor, key, value):
                        return True
                return False

            for k, v in self.match.items():
                if k.startswith('@'):
                    if not match_ancestors(k[1:], v):
                        return False
                else:
                    if not match_single(device, k, v):
                        return False

            if not self.filter_match(device):
                return False
        else:  # update
            if self.device.sys_path != device.sys_path:
                return False

        self.logger.debug(" found match: %s", self)

        if self.suggest and device.action in [None, 'add']:
            self.device = device
            self.suggest(self, *self.suggest_match(device))
            self.device = None
            return False

        if device.action in [None, 'add']:
            if self.avail:
                warnings.warn(f"udev device {device} is already available")
            self.device = device
        elif device.action in ['change', 'move']:
            self.device = device
        elif device.action in ['unbind', 'remove']:
            self.device = None

        self.avail = self.device is not None
        self.update()

        return True

    def update(self):
        pass

    @property
    def busnum(self):
        device = self._get_usb_device()
        if device:
            return int(device.properties.get('BUSNUM'))

        return None

    @property
    def devnum(self):
        device = self._get_usb_device()
        if device:
            return int(device.properties.get('DEVNUM'))

        return None

    def _get_usb_device(self):
        device = self.device
        if self.device is not None and (self.device.subsystem != 'usb'
                                        or self.device.device_type != 'usb_device'):
            device = self.device.find_parent('usb', 'usb_device')
        return device

    @property
    def path(self):
        device = self._get_usb_device()
        if device:
            return str(device.sys_name)

        return None

    @property
    def vendor_id(self):
        device = self._get_usb_device()
        if device:
            return int(device.properties.get('ID_VENDOR_ID'), 16)

        return None

    @property
    def model_id(self):
        device = self._get_usb_device()
        if device:
            return int(device.properties.get('ID_MODEL_ID'), 16)

        return None

    def read_attr(self, attribute):
        """read uncached attribute value"""
        if self.device is not None:
            self.device.attributes.unset(attribute)
            return self.device.attributes.get(attribute)

        return None


@target_factory.reg_resource
@attr.s(eq=False)
class USBSerialPort(USBResource, SerialPort):
    def __attrs_post_init__(self):
        self.match['SUBSYSTEM'] = 'tty'
        self.match['@SUBSYSTEM'] = 'usb'
        if self.port:
            warnings.warn(
                "USBSerialPort: The port attribute will be overwritten by udev.\n"
                "Please use udev matching as described in http://labgrid.readthedocs.io/en/latest/configuration.html#udev-matching"  # pylint: disable=line-too-long
            )
        super().__attrs_post_init__()

    def update(self):
        super().update()
        if self.device is not None:
            self.port = self.device.device_node
        else:
            self.port = None

@target_factory.reg_resource
@attr.s(eq=False)
class USBMassStorage(USBResource):
    def __attrs_post_init__(self):
        self.match['SUBSYSTEM'] = 'block'
        self.match['DEVTYPE'] = 'disk'
        self.match['@SUBSYSTEM'] = 'usb'
        super().__attrs_post_init__()

    # Overwrite the avail attribute with our internal property
    @property
    def avail(self):
        return self.path is not None

    # Forbid the USBResource super class to set the avail property
    @avail.setter
    def avail(self, prop):
        pass

    @property
    def path(self):
        if self.device is not None:
            return self.device.device_node

        return None

@target_factory.reg_resource
@attr.s(eq=False)
class IMXUSBLoader(USBResource):
    def filter_match(self, device):
        match = (device.properties.get('ID_VENDOR_ID'), device.properties.get('ID_MODEL_ID'))

        if match not in [("15a2", "0054"), ("15a2", "0061"),
                         ("15a2", "0063"), ("15a2", "0071"),
                         ("15a2", "007d"), ("15a2", "0076"),
                         ("15a2", "0080"), ("15a2", "003a"),
                         ("1fc9", "0128"), ("1fc9", "0126"),
                         ("1fc9", "012b"), ("1fc9", "0134"),
                         ("1fc9", "013e"), ("1fc9", "0146"),
                         ("1fc9", "014e"), ("1fc9", "0129"),
                         ("1fc9", "0159"), ("1fc9", "015d"),
                         ("1b67", "4fff"), ("0525", "b4a4"), # SPL
                         ("3016", "1001"),
                         ]:
            return False

        return super().filter_match(device)

@target_factory.reg_resource
@attr.s(eq=False)
class RKUSBLoader(USBResource):
    def filter_match(self, device):
        match = (device.properties.get('ID_VENDOR_ID'), device.properties.get('ID_MODEL_ID'))

        if match not in [("2207", "110a"),  # RV1108
                         ("2207", "110b"),  # RV1126
                         ("2207", "110c"),  # RV1106
                         ("2207", "110e"),  # RV1103B
                         ("2207", "110f"),  # RV1126B
                         ("2207", "300a"),  # RK3066
                         ("2207", "301a"),  # RK3036
                         ("2207", "310b"),  # RK3188
                         ("2207", "310c"),  # RK3128
                         ("2207", "320a"),  # RK3288
                         ("2207", "320b"),  # RK322X
                         ("2207", "320c"),  # RK3328
                         ("2207", "330a"),  # RK3368
                         ("2207", "330c"),  # RK3399
                         ("2207", "330d"),  # PX30
                         ("2207", "330e"),  # RK3308
                         ("2207", "350a"),  # RK3568
                         ("2207", "350b"),  # RK3588
                         ("2207", "350c"),  # RK3528
                         ("2207", "350d"),  # RK3562
                         ("2207", "350e"),  # RK3576
                         ("2207", "350f"),  # RK3506
                         ]:
            return False

        return super().filter_match(device)

@target_factory.reg_resource
@attr.s(eq=False)
class MXSUSBLoader(USBResource):
    def filter_match(self, device):
        match = (device.properties.get('ID_VENDOR_ID'), device.properties.get('ID_MODEL_ID'))

        if match not in [("066f", "3780"), ("15a2", "004f")]:
            return False

        return super().filter_match(device)

@target_factory.reg_resource
@attr.s(eq=False)
class AndroidUSBFastboot(USBResource):
    usb_vendor_id = attr.ib(default='1d6b', validator=attr.validators.instance_of(str))
    usb_product_id = attr.ib(default='0104', validator=attr.validators.instance_of(str))
    def filter_match(self, device):
        if device.properties.get("adb_user") == "yes":
            return True
        if device.properties.get('ID_VENDOR_ID') != self.usb_vendor_id:
            return False
        if device.properties.get('ID_MODEL_ID') != self.usb_product_id:
            return False
        return super().filter_match(device)

@target_factory.reg_resource
@attr.s(eq=False)
class AndroidFastboot(AndroidUSBFastboot):
    def __attrs_post_init__(self):
        warnings.warn("AndroidFastboot is deprecated, use AndroidUSBFastboot instead",
                      DeprecationWarning)
        super().__attrs_post_init__()

@target_factory.reg_resource
@attr.s(eq=False)
class DFUDevice(USBResource):
    def filter_match(self, device):
        if ':fe0102:' not in device.properties.get('ID_USB_INTERFACES', ''):
            return False

        return super().filter_match(device)

@target_factory.reg_resource
@attr.s(eq=False)
class USBNetworkInterface(USBResource, NetworkInterface):
    def __attrs_post_init__(self):
        self.match['SUBSYSTEM'] = 'net'
        self.match['@SUBSYSTEM'] = 'usb'
        if self.ifname:
            warnings.warn(
                "USBNetworkInterface: The ifname attribute will be overwritten by udev.\n"
                "Please use udev matching as described in http://labgrid.readthedocs.io/en/latest/configuration.html#udev-matching"  # pylint: disable=line-too-long
            )
        super().__attrs_post_init__()

    def update(self):
        super().update()
        if self.device is not None:
            self.ifname = self.device.properties.get('INTERFACE')
        else:
            self.ifname = None

    @property
    def if_state(self):
        value = self.read_attr('operstate')
        if value is not None:
            value = value.decode('ascii')
        return value

@target_factory.reg_resource
@attr.s(eq=False)
class AlteraUSBBlaster(USBResource):
    def filter_match(self, device):
        if device.properties.get('ID_VENDOR_ID') != "09fb":
            return False
        if device.properties.get('ID_MODEL_ID') not in ["6010", "6810"]:
            return False
        return super().filter_match(device)

@target_factory.reg_resource
@attr.s(eq=False)
class SigrokUSBDevice(USBResource):
    """The SigrokUSBDevice describes an attached sigrok device with driver and
    optional channel mapping, it is identified via usb using udev.

    This is used for devices which communicate over a custom USB protocol.

    Args:
        driver (str): driver to use with sigrok
        channels (str): a sigrok channel mapping as described in the sigrok-cli man page
        channel_group (str): a sigrok channel group as described in the sigrok-cli man page
    """
    driver = attr.ib(
        default=None,
        validator=attr.validators.instance_of(str)
    )
    channels = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str))
    )

    channel_group = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str))
    )

    def __attrs_post_init__(self):
        self.match['@SUBSYSTEM'] = 'usb'
        super().__attrs_post_init__()

@target_factory.reg_resource
@attr.s(eq=False)
class SigrokUSBSerialDevice(USBResource):
    """The SigrokUSBSerialDevice describes an attached sigrok device with driver and
    optional channel mapping, it is identified via usb using udev.

    This is used for devices which communicate over an emulated serial device.

    Args:
        driver (str): driver to use with sigrok
        channels (str): a sigrok channel mapping as described in the sigrok-cli man page
        channel_group (str): a sigrok channel group as described in the sigrok-cli man page
    """
    driver = attr.ib(
        default=None,
        validator=attr.validators.instance_of(str)
    )
    channels = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str))
    )
    channel_group = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str))
    )

    def __attrs_post_init__(self):
        self.match['SUBSYSTEM'] = 'tty'
        self.match['@SUBSYSTEM'] = 'usb'
        super().__attrs_post_init__()

    @property
    def path(self):
        if self.device is not None:
            return self.device.device_node

        return None

@target_factory.reg_resource
@attr.s(eq=False)
class USBSDWireDevice(USBResource):
    """The USBSDWireDevice describes an attached SDWire device,
    it is identified via USB using udev
    """

    disk_path = attr.ib(
        default=None,
        validator=attr.validators.optional(str)
    )

    def __attrs_post_init__(self):
        self.control_serial = None
        self.match['ID_VENDOR_ID'] = '04e8'
        self.match['ID_MODEL_ID'] = '6001'
        self.match['@ID_VENDOR_ID'] = '0424'
        self.match['@ID_MODEL_ID'] = '2640'
        super().__attrs_post_init__()

    # Overwrite the avail attribute with our internal property
    @property
    def avail(self):
        return bool(self.disk_path and self.control_serial)

    # Forbid the USBResource super class to set the avail property
    @avail.setter
    def avail(self, prop):
        pass

    # Overwrite the poll function. Only mark the SDWire as available if both
    # control_serial and disk_path are available.
    def poll(self):
        super().poll()
        if self.device is not None and not self.avail:
            for child in self.device.parent.children:
                if child.subsystem == 'block' and child.device_type == 'disk':
                    self.disk_path = child.device_node
            self.control_serial = self.device.properties.get('ID_SERIAL_SHORT')

    def update(self):
        super().update()
        if self.device is None:
            self.disk_path = None
            self.control_serial = None

    @property
    def path(self):
        return self.disk_path

@target_factory.reg_resource
@attr.s(eq=False)
class USBSDWire3Device(USBResource):
    """The USBSDWire3Device describes an attached SDWire3 device,
    it is identified via USB using udev
    """

    control_serial = attr.ib(
        default=None,
        validator=attr.validators.optional(str)
    )
    disk_path = attr.ib(
        default=None,
        validator=attr.validators.optional(str)
    )

    def __attrs_post_init__(self):
        self.match['ID_VENDOR_ID'] = '0bda'
        self.match['ID_MODEL_ID'] = '0316'
        self.match['@ID_VENDOR_ID'] = '1d6b'
        self.match['@ID_MODEL_ID'] = '0002'        
        super().__attrs_post_init__()

    # Overwrite the avail attribute with our internal property
    @property
    def avail(self):
        return bool(self.control_serial)

    # Forbid the USBResource super class to set the avail property
    @avail.setter
    def avail(self, prop):
        pass

    # Overwrite the poll function. Only mark the SDWire3 as available if both
    # paths are available.
    def poll(self):
        super().poll()
        if self.device is not None and not self.avail:
            for child in self.device.parent.children:
                if child.subsystem == 'block' and child.device_type == 'disk':
                    self.disk_path = child.device_node
            self.control_serial = self.device.properties.get('ID_SERIAL_SHORT')

    def update(self):
        super().update()
        if self.device is None:
            self.disk_path = None
            self.control_serial = None

    @property
    def path(self):
        return self.disk_path

@target_factory.reg_resource
@attr.s(eq=False)
class USBSDMuxDevice(USBResource):
    """The USBSDMuxDevice describes an attached USBSDMux device,
    it is identified via USB using udev
    """

    control_path = attr.ib(default=None)
    disk_path = attr.ib(default=None)

    def __attrs_post_init__(self):
        self.match['ID_VENDOR_ID'] = '0424'
        self.match['ID_MODEL_ID'] = '4041'
        super().__attrs_post_init__()

    # Overwrite the avail attribute with our internal property
    @property
    def avail(self):
        return bool(self.disk_path and self.control_path)

    # Forbid the USBResource super class to set the avail property
    @avail.setter
    def avail(self, prop):
        pass

    # Overwrite the poll function. Only mark the SDMux as available if both
    # paths are available.
    def poll(self):
        super().poll()
        if self.device is not None and not self.avail:
            for child in self.device.children:
                if child.subsystem == 'block' and child.device_type == 'disk':
                    self.disk_path = child.device_node
                elif child.subsystem == 'scsi_generic':
                    self.control_path = child.device_node

    def update(self):
        super().update()
        if self.device is None:
            self.control_path = None
            self.disk_path = None

    @property
    def path(self):
        return self.disk_path


@target_factory.reg_resource
@attr.s(eq=False)
class USBHub(USBResource):
    """The USBHub describes a USB hub.

    This is mainly useful to monitor if all expected hubs are detected.
    """
    def __attrs_post_init__(self):
        self.match['DEVTYPE'] = 'usb_interface'
        self.match['DRIVER'] = 'hub'
        super().__attrs_post_init__()


@target_factory.reg_resource
@attr.s(eq=False)
class USBPowerPort(USBResource):
    """The USBPowerPort describes a single port on an USB hub which supports
    power control.

    Args:
        index (int): index of the downstream port on the USB hub
    """
    index = attr.ib(default=None, validator=attr.validators.instance_of(int))
    def __attrs_post_init__(self):
        self.match['DEVTYPE'] = 'usb_interface'
        self.match['DRIVER'] = 'hub'
        super().__attrs_post_init__()

@target_factory.reg_resource
@attr.s(eq=False)
class SiSPMPowerPort(USBResource):
    """This resource describes a SiS-PM (Silver Shield PM) power port.

    Args:
        index (int): port index
    """
    index = attr.ib(default=0, validator=attr.validators.instance_of(int))

    def filter_match(self, device):
        match = (device.properties.get('ID_VENDOR_ID'), device.properties.get('ID_MODEL_ID'))

        if match not in [
                ("04b4", "fd10"), ("04b4", "fd11"),
                ("04b4", "fd12"), ("04b4", "fd13"),
                ("04b4", "fd15")]:
            return False

        return super().filter_match(device)

@target_factory.reg_resource
@attr.s(eq=False)
class USBVideo(USBResource):
    def filter_match(self, device):
        capabilities = device.properties.get('ID_V4L_CAPABILITIES').split(':')

        if 'capture' not in capabilities:
            return False

        return super().filter_match(device)

    def __attrs_post_init__(self):
        self.match['SUBSYSTEM'] = 'video4linux'
        self.match['@SUBSYSTEM'] = 'usb'
        super().__attrs_post_init__()

    @property
    def path(self):
        if self.device is not None:
            return self.device.device_node

        return None

@target_factory.reg_resource
@attr.s(eq=False)
class USBAudioInput(USBResource):
    # the ALSA PCM device number (as in hw:CARD=<card>,DEV=<index>)
    index = attr.ib(default=0, validator=attr.validators.instance_of(int))

    def filter_match(self, device):
        suffix = f"D{self.index}c"

        if not device.sys_name.endswith(suffix):
            return False

        return super().filter_match(device)

    def __attrs_post_init__(self):
        self.match['SUBSYSTEM'] = 'sound'
        self.match['@SUBSYSTEM'] = 'usb'
        super().__attrs_post_init__()

    @property
    def path(self):
        "the device node (for example /dev/snd/pcmC3D0c)"
        if self.device is not None:
            return self.device.device_node

        return None

    @property
    def alsa_name(self):
        "the ALSA device name (for example hw:CARD=3,DEV=0)"
        if self.device is not None:
            return f"hw:CARD={self.device.parent.attributes.asint('number')},DEV={self.index}"

        return None

@target_factory.reg_resource
@attr.s(eq=False)
class USBTMC(USBResource):
    def __attrs_post_init__(self):
        self.match['SUBSYSTEM'] = 'usbmisc'
        self.match['@DRIVER'] = 'usbtmc'
        self.match['@SUBSYSTEM'] = 'usb'
        super().__attrs_post_init__()

    @property
    def path(self):
        if self.device is not None:
            return self.device.device_node

        return None

@target_factory.reg_resource
@attr.s(eq=False)
class DeditecRelais8(USBResource):
    index = attr.ib(default=None, validator=attr.validators.instance_of(int))
    invert = attr.ib(default=False, validator=attr.validators.instance_of(bool))

    def __attrs_post_init__(self):
        self.match['ID_VENDOR'] = 'DEDITEC'
        # Deditec does not set proper model/vendor IDs, so we use ID_MODEL instead
        self.match['ID_MODEL'] = 'DEDITEC_USB-OPT_REL-8'
        super().__attrs_post_init__()


@target_factory.reg_resource
@attr.s(eq=False)
class HIDRelay(USBResource):
    index = attr.ib(default=1, validator=attr.validators.instance_of(int))
    invert = attr.ib(default=False, validator=attr.validators.instance_of(bool))

    def filter_match(self, device):
        match = (device.properties.get('ID_VENDOR_ID'), device.properties.get('ID_MODEL_ID'))

        if match not in [("16c0", "05df"),  # dcttech USBRelay2
                         ("5131", "2007"),  # LC-US8
                         ]:
            return False

        return super().filter_match(device)

@target_factory.reg_resource
@attr.s(eq=False)
class USBFlashableDevice(USBResource):
    @property
    def devnode(self):
        if self.device is not None:
            return self.device.device_node
        return None

@target_factory.reg_resource
@attr.s(eq=False)
class LXAUSBMux(USBResource):
    """The LXAUSBMux describes an attached USBMux device,
    it is identified via USB using udev.
    """

    def __attrs_post_init__(self):
        self.match['ID_VENDOR_ID'] = '33f7'
        self.match['ID_MODEL_ID'] = '0001'
        super().__attrs_post_init__()

@target_factory.reg_resource
@attr.s(eq=False)
class USBDebugger(USBResource):
    def filter_match(self, device):
        match = (device.properties.get('ID_VENDOR_ID'), device.properties.get('ID_MODEL_ID'))

        if match not in [("0403", "6010"),  # FT2232C/D/H Dual UART/FIFO IC
                         ("0403", "6014"),  # FT232HL/Q
                         ("0483", "3748"),  # STLINK-V2
                         ("0483", "374b"),  # STLINK-V3
                         ("0483", "374e"),  # STLINK-V3
                         ("0483", "374f"),  # STLINK-V3
                         ("0483", "3754"),  # STLINK-V3
                         ("04b4", "f155"),  # KitProg3 CMSIS-DAP
                         ("15ba", "0003"),  # Olimex ARM-USB-OCD
                         ("15ba", "002b"),  # Olimex ARM-USB-OCD-H
                         ("15ba", "0004"),  # Olimex ARM-USB-TINY
                         ("15ba", "002a"),  # Olimex ARM-USB-TINY-H
                         # SEGGER J-Link IDs extracted from udev rules as shipped by the j-link driver package
                         # (https://www.segger.com/downloads/jlink/JLink_Linux_x86_64.tgz)
                         ("1366", "0101"),  # SEGGER J-Link
                         ("1366", "0102"),  # SEGGER J-Link
                         ("1366", "0103"),  # SEGGER J-Link
                         ("1366", "0104"),  # SEGGER J-Link
                         ("1366", "0105"),  # SEGGER J-Link
                         ("1366", "0107"),  # SEGGER J-Link
                         ("1366", "0108"),  # SEGGER J-Link
                         ("1366", "1001"),  # SEGGER J-Link
                         ("1366", "1002"),  # SEGGER J-Link
                         ("1366", "1003"),  # SEGGER J-Link
                         ("1366", "1004"),  # SEGGER J-Link
                         ("1366", "1005"),  # SEGGER J-Link
                         ("1366", "1006"),  # SEGGER J-Link
                         ("1366", "1007"),  # SEGGER J-Link
                         ("1366", "1008"),  # SEGGER J-Link
                         ("1366", "1009"),  # SEGGER J-Link
                         ("1366", "100a"),  # SEGGER J-Link
                         ("1366", "100b"),  # SEGGER J-Link
                         ("1366", "100c"),  # SEGGER J-Link
                         ("1366", "100d"),  # SEGGER J-Link
                         ("1366", "100e"),  # SEGGER J-Link
                         ("1366", "100f"),  # SEGGER J-Link
                         ("1366", "1010"),  # SEGGER J-Link
                         ("1366", "1011"),  # SEGGER J-Link
                         ("1366", "1012"),  # SEGGER J-Link
                         ("1366", "1013"),  # SEGGER J-Link
                         ("1366", "1014"),  # SEGGER J-Link
                         ("1366", "1015"),  # SEGGER J-Link
                         ("1366", "1016"),  # SEGGER J-Link
                         ("1366", "1017"),  # SEGGER J-Link
                         ("1366", "1018"),  # SEGGER J-Link
                         ("1366", "1019"),  # SEGGER J-Link
                         ("1366", "101a"),  # SEGGER J-Link
                         ("1366", "101b"),  # SEGGER J-Link
                         ("1366", "101c"),  # SEGGER J-Link
                         ("1366", "101d"),  # SEGGER J-Link
                         ("1366", "101e"),  # SEGGER J-Link
                         ("1366", "101f"),  # SEGGER J-Link
                         ("1366", "1020"),  # SEGGER J-Link
                         ("1366", "1021"),  # SEGGER J-Link
                         ("1366", "1022"),  # SEGGER J-Link
                         ("1366", "1023"),  # SEGGER J-Link
                         ("1366", "1024"),  # SEGGER J-Link
                         ("1366", "1025"),  # SEGGER J-Link
                         ("1366", "1026"),  # SEGGER J-Link
                         ("1366", "1027"),  # SEGGER J-Link
                         ("1366", "1028"),  # SEGGER J-Link
                         ("1366", "1029"),  # SEGGER J-Link
                         ("1366", "102a"),  # SEGGER J-Link
                         ("1366", "102b"),  # SEGGER J-Link
                         ("1366", "102c"),  # SEGGER J-Link
                         ("1366", "102d"),  # SEGGER J-Link
                         ("1366", "102e"),  # SEGGER J-Link
                         ("1366", "102f"),  # SEGGER J-Link
                         ("1366", "1050"),  # SEGGER J-Link
                         ("1366", "1051"),  # SEGGER J-Link
                         ("1366", "1052"),  # SEGGER J-Link
                         ("1366", "1053"),  # SEGGER J-Link
                         ("1366", "1054"),  # SEGGER J-Link
                         ("1366", "1055"),  # SEGGER J-Link
                         ("1366", "1056"),  # SEGGER J-Link
                         ("1366", "1057"),  # SEGGER J-Link
                         ("1366", "1058"),  # SEGGER J-Link
                         ("1366", "1059"),  # SEGGER J-Link
                         ("1366", "105a"),  # SEGGER J-Link
                         ("1366", "105b"),  # SEGGER J-Link
                         ("1366", "105c"),  # SEGGER J-Link
                         ("1366", "105d"),  # SEGGER J-Link
                         ("1366", "105e"),  # SEGGER J-Link
                         ("1366", "105f"),  # SEGGER J-Link
                         ("1366", "1060"),  # SEGGER J-Link
                         ("1366", "1061"),  # SEGGER J-Link
                         ("1366", "1062"),  # SEGGER J-Link
                         ("1366", "1063"),  # SEGGER J-Link
                         ("1366", "1064"),  # SEGGER J-Link
                         ("1366", "1065"),  # SEGGER J-Link
                         ("1366", "1066"),  # SEGGER J-Link
                         ("1366", "1067"),  # SEGGER J-Link
                         ("1366", "1068"),  # SEGGER J-Link
                         ("1366", "1069"),  # SEGGER J-Link
                         ("1366", "106a"),  # SEGGER J-Link
                         ("1366", "106b"),  # SEGGER J-Link
                         ("1366", "106c"),  # SEGGER J-Link
                         ("1366", "106d"),  # SEGGER J-Link
                         ("1366", "106e"),  # SEGGER J-Link
                         ("1366", "106f"),  # SEGGER J-Link
                         ("064b", "0617"),  # Analog Devices ICE-1000 Emulator
                         ("064b", "2500"),  # Analog Devices ICE-1500 Emulator
                         ("064b", "0283"),  # Analog Devices ICE-2000 Emulator
                         ("064b", "2503"),  # Analog Devices Onboard Debug Agent
                         ("064b", "2504"),  # Analog Devices Onboard Debug Agent
                         ("064b", "2507"),  # Analog Devices Onboard Debug Agent
                         ("064b", "2508"),  # Analog Devices Onboard Debug Agent
                         ("064b", "250A"),  # Analog Devices Onboard Debug Agent
                         ]:
            return False

        return super().filter_match(device)

@target_factory.reg_resource
@attr.s(eq=False)
class MatchedSysfsGPIO(USBResource):
    """The MatchedSysfsGPIO described a SysfsGPIO matched by Udev

    Args:
        pin (int): gpio pin number within the matched gpiochip."""
    pin = attr.ib(default=None, validator=attr.validators.instance_of(int))
    index = None

    def __attrs_post_init__(self):
        self.match['SUBSYSTEM'] = 'gpio'
        super().__attrs_post_init__()

    def filter_match(self, device):
        # Filter out the char device
        if device.properties.get('DEVNAME') is not None:
            return False
        return super().filter_match(device)

    def update(self):
        super().update()
        if self.device is not None:
            if self.pin >= int(self.read_attr('ngpio')):
                raise ValueError("MatchedSysfsGPIO pin out of bound")
            self.index = int(self.read_attr('base')) + self.pin
        else:
            self.index = None



#!/usr/bin/env python3

import json
import os
import signal
import sys
import base64
import types
import socket
import pickle

def b2s(b):
    return base64.b85encode(b).decode('ascii')

def s2b(s):
    return base64.b85decode(s.encode('ascii'))


# Similar to b2s and s2b, py2s and s2py will convert python objects to strings
# using pickle. This is important when exact python types need to be passed to
# the wrappers, for example tuples.
# Only use these to pass parameters from the wrapper to the agent, as
# unpickling is not secure with untrusted data.

def py2s(o):
    return b2s(pickle.dumps(o))

def s2py(s):
    return pickle.loads(s2b(s))

class Agent:
    def __init__(self):
        self.methods = {}
        self.register('load', self.load)
        self.register('list', self.list)

        # use real stdin/stdout
        self.stdin = sys.stdin
        self.stdout = sys.stdout

        # use stderr for normal prints
        sys.stdout = sys.stderr

        self.fdpass = None
        if fdpass_env := os.environ.get("LG_FDPASS"):
            self.fdpass = socket.socket(fileno=int(fdpass_env))

    def send(self, data):
        self.stdout.write(json.dumps(data)+'\n')
        self.stdout.flush()

    def register(self, name, func):
        assert name not in self.methods
        self.methods[name] = func

    def load(self, name, source):
        module = types.ModuleType(name)
        exec(compile(source, f'<loaded {name}>', 'exec'), module.__dict__)
        for k, v in module.methods.items():  # pylint: disable=no-member
            self.register(f'{name}.{k}', v)

    def list(self):
        return list(self.methods.keys())

    def run(self):
        for line in self.stdin:
            if not line:
                continue

            try:
                request = json.loads(line)
            except json.JSONDecodeError:
                self.send({'error': f'request parsing failed for {repr(line)}'})
                break

            if request.get('close', False):
                break

            name = request['method']
            args = request['args']
            kwargs = request['kwargs']
            try:
                response = self.methods[name](*args, **kwargs)
                # check if the method returned a file descriptor
                if isinstance(response, tuple) and len(response) == 2 and hasattr(response[1], 'fileno'):
                    try:
                        if self.fdpass is None:
                            self.send({'error': 'cannot pass returned FD without LG_FDPASS'})
                            break
                        socket.send_fds(self.fdpass, [b"\0"], (response[1].fileno(),))
                        self.send({'result': response[0], 'fdpass': True})
                    finally:
                        response[1].close()
                else:
                    self.send({'result': response})
            except Exception as e:  # pylint: disable=broad-except
                import traceback
                try:
                    tb = [list(x) for x in traceback.extract_tb(sys.exc_info()[2])]
                except:
                    tb = None
                self.send({'exception': repr(e), 'tb': tb})

def handle_test(*args, **kwargs):  # pylint: disable=unused-argument
    return args[::-1]

def handle_test_fd():
    fd = os.fdopen(os.memfd_create("test_fd"))
    return ("dummy", fd)

def handle_error(message):
    raise ValueError(message)

def handle_usbtmc(index, cmd, read=False):
    assert isinstance(index, int)
    cmd = s2b(cmd)
    fd = os.open(f'/dev/usbtmc{index}', os.O_RDWR)
    os.write(fd, cmd)
    if not read:
        os.close(fd)
        return None
    data = []
    while True:
        data.append(os.read(fd, 4096))
        if len(data[-1]) < 4096:
            break
    os.close(fd)
    return b2s(b''.join(data))

def main():
    signal.signal(signal.SIGINT, signal.SIG_IGN)

    a = Agent()
    a.register('test', handle_test)
    a.register('test_fd', handle_test_fd)
    a.register('error', handle_error)
    a.register('usbtmc', handle_usbtmc)
    a.run()

if __name__ == "__main__":
    main()



import os
import shutil
import tempfile
import time
from importlib import import_module
from urllib.parse import urlparse
from urllib.request import urlopen

import attr

from ..factory import target_factory
from ..protocol import ConsoleProtocol, PowerProtocol
from ..step import step
from .common import Driver
from .consoleexpectmixin import ConsoleExpectMixin
from .exception import ExecutionError
from .powerdriver import PowerResetMixin


def _get_laam():
    try:
        return import_module('laam')
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            "laam package not found, install labgrid[laa]"
        )


@target_factory.reg_driver
@attr.s(eq=False)
class LAASerialDriver(ConsoleExpectMixin, Driver, ConsoleProtocol):
    """Driver for serial console via LAA WebSocket bridge"""
    bindings = {"port": "LAASerialPort", }

    txdelay = attr.ib(default=0.0, validator=attr.validators.instance_of(float))
    timeout = attr.ib(default=3.0, validator=attr.validators.instance_of(float))

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self._laam = _get_laam()
        self._laa = None
        self._conn = None

    def on_activate(self):
        self._laa = self._laam.LAA(self.port.laa_identity)
        self._conn = self._laa.serials.connect_pexpect(self.port.serial_name)

    def on_deactivate(self):
        if self._conn is not None:
            self._conn.close()
            self._conn = None
        self._laa = None

    def _read(self, size: int = 1, timeout: float = 0.0, max_size: int = None):
        # ConnectPexpect.read_nonblocking() returns str, labgrid expects bytes
        data = self._conn.read_nonblocking(size, timeout)
        return data.encode("utf-8", errors="replace")

    def _write(self, data: bytes):
        return self._conn.send(data)


@target_factory.reg_driver
@attr.s(eq=False)
class LAAPowerDriver(Driver, PowerResetMixin, PowerProtocol):
    """Driver for DUT power control via LAA"""
    bindings = {"port": "LAAPowerPort", }

    delay = attr.ib(default=2.0, validator=attr.validators.instance_of(float))

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self._laam = _get_laam()
        self._laa = None

    def on_activate(self):
        self._laa = self._laam.LAA(self.port.laa_identity)

    def on_deactivate(self):
        self._laa = None

    def _execute_sequence(self, sequence):
        for vbus, state in sequence:
            try:
                self._laa.laacli.power(vbus, state)
            except self._laam.exceptions.LAAError as e:
                raise ExecutionError(
                    f"LAA power command failed ({vbus} {state}): {e}"
                ) from e

    @Driver.check_active
    @step()
    def on(self):
        self._execute_sequence(self.port.power_on)

    @Driver.check_active
    @step()
    def off(self):
        self._execute_sequence(self.port.power_off)

    @Driver.check_active
    @step()
    def cycle(self):
        if self.port.power_cycle is not None:
            self._execute_sequence(self.port.power_cycle)
        else:
            self.off()
            time.sleep(self.delay)
            self.on()


@target_factory.reg_driver
@attr.s(eq=False)
class LAAUSBGadgetMassStorageDriver(Driver):
    """Driver for USB gadget mass storage control via LAA"""
    bindings = {"port": "LAAUSBGadgetMassStorage", }

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self._laam = _get_laam()
        self._laa = None

    def on_activate(self):
        self._laa = self._laam.LAA(self.port.laa_identity)

    def on_deactivate(self):
        self._laa = None

    @Driver.check_active
    @step()
    def on(self):
        try:
            self._laa.laacli.usbg_ms("on", self.port.image)
        except self._laam.exceptions.LAAError as e:
            raise ExecutionError(f"LAA usbg-ms on failed: {e}") from e

    @Driver.check_active
    @step()
    def off(self):
        try:
            self._laa.laacli.usbg_ms("off", "")
        except self._laam.exceptions.LAAError as e:
            raise ExecutionError(f"LAA usbg-ms off failed: {e}") from e


@target_factory.reg_driver
@attr.s(eq=False)
class LAAUSBDriver(Driver):
    """Driver for USB port control via LAA"""
    bindings = {"port": "LAAUSBPort", }

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self._laam = _get_laam()
        self._laa = None

    def on_activate(self):
        self._laa = self._laam.LAA(self.port.laa_identity)

    def on_deactivate(self):
        self._laa = None

    @Driver.check_active
    @step()
    def on(self):
        for port in self.port.usb_ports:
            try:
                self._laa.laacli.usb(port, "on")
            except self._laam.exceptions.LAAError as e:
                raise ExecutionError(
                    f"LAA USB command failed (port {port} on): {e}"
                ) from e

    @Driver.check_active
    @step()
    def off(self):
        for port in self.port.usb_ports:
            try:
                self._laa.laacli.usb(port, "off")
            except self._laam.exceptions.LAAError as e:
                raise ExecutionError(
                    f"LAA USB command failed (port {port} off): {e}"
                ) from e


@target_factory.reg_driver
@attr.s(eq=False)
class LAAButtonDriver(Driver):
    """Driver for virtual button control via LAA"""
    bindings = {"port": "LAAButtonPort", }

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self._laam = _get_laam()
        self._laa = None

    def on_activate(self):
        self._laa = self._laam.LAA(self.port.laa_identity)

    def on_deactivate(self):
        self._laa = None

    @Driver.check_active
    @step()
    def press(self, button):
        if button not in self.port.buttons:
            raise ExecutionError(
                f"Button '{button}' not in resource buttons {self.port.buttons}"
            )
        try:
            self._laa.laacli.button(button, "on")
        except self._laam.exceptions.LAAError as e:
            raise ExecutionError(f"LAA button command failed ({button} on): {e}") from e

    @Driver.check_active
    @step()
    def release(self, button):
        if button not in self.port.buttons:
            raise ExecutionError(
                f"Button '{button}' not in resource buttons {self.port.buttons}"
            )
        try:
            self._laa.laacli.button(button, "off")
        except self._laam.exceptions.LAAError as e:
            raise ExecutionError(
                f"LAA button command failed ({button} off): {e}"
            ) from e


@target_factory.reg_driver
@attr.s(eq=False)
class LAALedDriver(Driver):
    """Driver for LED control via LAA"""
    bindings = {"port": "LAALed", }

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self._laam = _get_laam()
        self._laa = None

    def on_activate(self):
        self._laa = self._laam.LAA(self.port.laa_identity)

    def on_deactivate(self):
        self._laa = None

    @Driver.check_active
    @step()
    def on(self):
        try:
            self._laa.laacli.led("on")
        except self._laam.exceptions.LAAError as e:
            raise ExecutionError(f"LAA LED on failed: {e}") from e

    @Driver.check_active
    @step()
    def off(self):
        try:
            self._laa.laacli.led("off")
        except self._laam.exceptions.LAAError as e:
            raise ExecutionError(f"LAA LED off failed: {e}") from e


@target_factory.reg_driver
@attr.s(eq=False)
class LAATempDriver(Driver):
    """Driver for temperature sensor reading via LAA"""
    bindings = {"port": "LAATempSensor", }

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self._laam = _get_laam()
        self._laa = None

    def on_activate(self):
        self._laa = self._laam.LAA(self.port.laa_identity)

    def on_deactivate(self):
        self._laa = None

    @Driver.check_active
    @step()
    def get_temp(self, probe):
        try:
            return self._laa.laacli.temp(probe)
        except self._laam.exceptions.LAAError as e:
            raise ExecutionError(f"LAA temp command failed ({probe}): {e}") from e


@target_factory.reg_driver
@attr.s(eq=False)
class LAAWattDriver(Driver):
    """Driver for power measurement via LAA"""
    bindings = {"port": "LAAWattMeter", }

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self._laam = _get_laam()
        self._laa = None

    def on_activate(self):
        self._laa = self._laam.LAA(self.port.laa_identity)

    def on_deactivate(self):
        self._laa = None

    @Driver.check_active
    @step()
    def get_watts(self, vbus):
        try:
            return self._laa.laacli.watt(vbus)
        except self._laam.exceptions.LAAError as e:
            raise ExecutionError(f"LAA watt command failed ({vbus}): {e}") from e


@target_factory.reg_driver
@attr.s(eq=False)
class LAAProviderDriver(Driver):
    """Driver for uploading files to LAA file storage.

    Files are served via TFTP to the DUT. Source can be a local path
    or a URL."""
    bindings = {"provider": "LAAProvider", }

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self._laam = _get_laam()
        self._laa = None

    def on_activate(self):
        self._laa = self._laam.LAA(self.provider.laa_identity)

    def on_deactivate(self):
        self._laa = None

    @Driver.check_active
    @step(args=['source'], result=True)
    def stage(self, source):
        """Upload a file to the LAA. Source can be a local path or URL.

        Returns the filename as stored on the LAA."""
        is_url = source.startswith("http://") or source.startswith("https://")

        if is_url:
            name = os.path.basename(urlparse(source).path)
            with tempfile.NamedTemporaryFile(delete=False) as tmp:
                tmp_path = tmp.name
            try:
                with urlopen(source) as r, open(tmp_path, "wb") as f:  # noqa: S310
                    shutil.copyfileobj(r, f)
                self._laa.files.push(name, tmp_path)
            finally:
                os.unlink(tmp_path)
        else:
            name = os.path.basename(source)
            if not os.path.exists(source):
                raise ExecutionError(f"file not found: {source}")
            self._laa.files.push(name, source)

        return name

    @Driver.check_active
    @step(result=True)
    def list(self):
        """Return list of files on the LAA."""
        return self._laa.files.list()

    @Driver.check_active
    @step(args=['name'])
    def remove(self, name):
        """Remove a file from the LAA."""
        try:
            self._laa.files.remove(name)
        except self._laam.exceptions.LAAError as e:
            raise ExecutionError(f"LAA file remove failed ({name}): {e}") from e



import pytest

from labgrid.util import Timeout


class TestTimeout:
    def test_create(self):
        t = Timeout()
        assert (isinstance(t, Timeout))
        t = Timeout(5.0)
        assert (isinstance(t, Timeout))
        with pytest.raises(TypeError):
            t = Timeout(10)
        with pytest.raises(ValueError):
            t = Timeout(-1.0)

    def test_expire(self, mocker):
        m = mocker.patch('time.monotonic')
        m.return_value = 0.0

        t = Timeout(5.0)
        assert not t.expired
        assert t.remaining == 5.0

        m.return_value += 3.0
        assert not t.expired
        assert t.remaining == 2.0

        m.return_value += 3.0
        assert t.expired
        assert t.remaining == 0.0



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# labgrid documentation build configuration file, created by
# sphinx-quickstart on Mon Feb 20 10:00:00 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.insert(0, os.path.abspath('..'))

from importlib import metadata

# Import read_the_docs theme
import sphinx_rtd_theme

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['sphinx.ext.autodoc',
              'sphinx.ext.doctest',
              'sphinx.ext.napoleon',
              'sphinx.ext.coverage',
              'sphinx.ext.viewcode',
              'sphinxcontrib.autoprogram',
              'sphinx.ext.autosectionlabel',
              'sphinx_rtd_theme']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['.templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'labgrid'
copyright = '2016-2025 Pengutronix, Jan Luebbe and Rouven Czerwinski'
author = 'Jan Luebbe, Rouven Czerwinski'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The full version, including alpha/beta/rc tags.
release = metadata.version('labgrid')
# The short X.Y version.
version = '.'.join(release.split('.')[:2])

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['.build', 'Thumbs.db', '.DS_Store', 'RELEASE.rst']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = 'images/labgrid_logo_outline.svg'

# The name of an image file (within the static path) to use as favicon of the
# docs.
html_favicon = 'images/favicon.png'


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'labgriddoc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'labgrid.tex', 'labgrid Documentation',
     'Jan Luebbe, Rouven Czerwinski', 'manual'),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'labgrid', 'labgrid Documentation',
     [author], 1),
    ('man/client', 'labgrid-client', 'labgrid\'s client interface to control boards',
     [author], 1),
    ('man/coordinator', 'labgrid-coordinator', 'managing labgrid resources and places',
     [author], 1),
    ('man/device-config', 'labgrid-device-config', 'test configuration files',
     [author], 5),
    ('man/exporter', 'labgrid-exporter', 'interface to control boards',
     [author], 1),
    ('man/pytest', 'labgrid-pytest', 'labgrid integration for pytest',
     [author], 7),
    ('man/suggest', 'labgrid-suggest', 'generator for YAML config files',
     [author], 1),
]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'labgrid', 'labgrid Documentation',
     author, 'labgrid', 'One line description of project.',
     'Miscellaneous'),
]


# -- Options for autodoc --------------------------------------------------

autodoc_member_order = 'bysource'
# TODO: __annotate_func__ can be dropped once
# https://github.com/sphinx-doc/sphinx/commit/78148440556d3e5962771c1025c85b253ccd73c8
# is part of a sphinx release
autodoc_default_options = {
        'special-members': True,
        'exclude-members': '__annotate_func__',
}
autodoc_mock_imports = ['onewire',
                        'gi',
                        'gi.repository',
                        'vxi11',
                        'pysnmp',
                        'kasa',
                        'kasa.iot',
                        ]

# -- Options for autosection ----------------------------------------------
autosectionlabel_prefix_document = True


from unittest.mock import Mock
for mod in autodoc_mock_imports:
    sys.modules[mod] = Mock()

def run_apidoc(app):
    from sphinx.ext.apidoc import main
    module = os.path.abspath(os.path.join(app.srcdir, '..', 'labgrid'))
    output = os.path.abspath(os.path.join(app.srcdir, 'modules'))
    cmd = [module, '-a', '-M', '-H', 'Modules', '-o', output]
    main(cmd)

def setup(app):
    # Make version and date stable when generating manpages as they will be tracked in git
    if app.outdir.parts[-1] == "man":
        app.config.version = ""
        app.config.today = "2025"
    app.connect('builder-inited', run_apidoc)
    app.connect('doctree-read', write_literal_blocks)

# -- Options for doctest --------------------------------------------------

doctest_global_setup = '''
import os
import shutil
from unittest.mock import Mock, patch

doctest_dir = '.build/doctest'

if not os.getcwd().endswith(doctest_dir):
    os.chdir(doctest_dir)
'''

doctest_global_cleanup = '''
if os.getcwd().endswith(doctest_dir):
    os.chdir('../..')
'''

def write_literal_blocks(app, doctree):
    """
    Writes named literal blocks to a file with that respective name in the temporary doctest build
    directory. This allows doctest to test code snippets referring to these files as-is.
    """
    blocks = doctree.traverse(
        condition=lambda node: node.tagname == 'literal_block'
    )

    for block in blocks:
        name = '_'.join(block['names'])

        if not name:
            continue

        out_path = os.path.join(app.outdir, name)

        if os.path.exists(out_path):
            raise Exception(f'literal block name "{name}" used multiple times')

        with open(out_path, 'w') as f:
            f.write(block.astext())



import argparse
import logging
import sys
import textwrap
import time

from .udev import (
    USBSerialPort,
    USBMassStorage,
    USBTMC,
    USBVideo,
    IMXUSBLoader,
    AndroidUSBFastboot,
    DFUDevice,
    USBSDMuxDevice,
    USBSDWireDevice,
    USBSDWire3Device,
    AlteraUSBBlaster,
    RKUSBLoader,
    USBNetworkInterface,
    SiSPMPowerPort,
    USBAudioInput,
    LXAUSBMux,
    HIDRelay,
    USBDebugger,
    USBPowerPort,
    MatchedSysfsGPIO
)
from ..util import dump


class Suggester:
    def __init__(self, args):
        self.resources = []
        self.log = logging.getLogger("suggester")

        args = {
            'target': None,
            'name': None,
            'suggest': self.suggest_callback,
        }

        self.resources.append(USBSerialPort(**args))
        self.resources.append(USBTMC(**args))
        self.resources.append(USBVideo(**args))
        self.resources.append(IMXUSBLoader(**args))
        self.resources.append(AndroidUSBFastboot(**args))
        self.resources.append(DFUDevice(**args))
        self.resources.append(USBMassStorage(**args))
        self.resources.append(USBSDMuxDevice(**args))
        self.resources.append(USBSDWireDevice(**args))
        self.resources.append(USBSDWire3Device(**args))
        self.resources.append(AlteraUSBBlaster(**args))
        self.resources.append(RKUSBLoader(**args))
        self.resources.append(USBNetworkInterface(**args))
        self.resources.append(SiSPMPowerPort(**args))
        self.resources.append(USBAudioInput(**args))
        self.resources.append(LXAUSBMux(**args))
        self.resources.append(HIDRelay(**args))
        self.resources.append(USBDebugger(**args))
        self.resources.append(USBPowerPort(**args, index=0))
        self.resources.append(MatchedSysfsGPIO(**args, pin=0))

    def suggest_callback(self, resource, meta, suggestions):
        cls = type(resource).__name__
        if resource.device.action == 'add':
            print("=== added device ===")
        else:
            print("=== existing device ===")
        print(f"  {cls} for {resource.device.device_path}")

        if meta:
            print("  === device properties ===")
        for k, v in meta.items():
            print(f"  {k}: {v}")
        if not suggestions:
            print("  === no suggested matches found ===")
            print()
            return

        print("  === suggested matches ===")
        for i, suggestion in enumerate(suggestions):
            if i:
                print("  ---")
            print(textwrap.indent(
                dump({cls: {"match": suggestion}}).strip(),
                '  ',
            ))
            if cls == 'USBPowerPort':
                print('    index: ?')
            if cls == 'MatchedSysfsGPIO':
                print('    pin: ?')
        print("  ---")
        print()

    def run(self):
        while True:
            managers = {r.get_managed_parent().manager for r in self.resources}
            for manager in managers:
                manager.poll()
            time.sleep(0.1)


def main():
    logging.basicConfig(
        level=logging.INFO,
        format='%(levelname)7s %(name)-20s %(message)s',
        stream=sys.stderr,
    )

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-d',
        '--debug',
        action='store_true',
        default=False,
        help="enable debug mode"
    )

    args = parser.parse_args()

    if args.debug:
        logging.getLogger().setLevel(logging.DEBUG)

    suggester = Suggester(args)
    try:
        suggester.run()
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    main()



import pytest

from labgrid.driver import FileDigitalOutputDriver


def test_filedigital_instance(target, mocker):
    m_isfile = mocker.patch("os.path.isfile", return_value=True)
    d = FileDigitalOutputDriver(target, name=None, filepath="/dev/null")
    target.activate(d)
    assert isinstance(d, FileDigitalOutputDriver)
    m_isfile.assert_called_once_with("/dev/null")


def test_filedigital_set(target, mocker):
    m_isfile = mocker.patch("os.path.isfile", return_value=True)
    m_open = mocker.patch("builtins.open", mocker.mock_open())
    d = FileDigitalOutputDriver(target, name=None, filepath="/dev/null")
    target.activate(d)
    d.set(True)
    m_isfile.assert_called_once_with("/dev/null")
    m_open.assert_called_once_with("/dev/null", "w")
    m_open.return_value.write.assert_called_once_with("1\n")


def test_filedigital_get(target, mocker):
    m_isfile = mocker.patch("os.path.isfile", return_value=True)
    m_open = mocker.patch("builtins.open", mocker.mock_open(read_data="1\n"))
    d = FileDigitalOutputDriver(target, name=None, filepath="/dev/null")
    target.activate(d)
    assert d.get() is True
    m_isfile.assert_called_once_with("/dev/null")
    m_open.assert_called_once_with("/dev/null")
    m_open.return_value.read.assert_called_once()



from datetime import datetime


def test_hwclock_rate(command):
    """Test that the hardware clock rate is not too inaccurate."""
    result = command.run_check("hwclock -c | head -n 3")
    hw_time, sys_time, freq_offset_ppm, tick = result[-1].strip().split()
    assert abs(int(freq_offset_ppm)) < 1000


def test_hwclock_value(command):
    """Test that the hardware clock has the correct time.

    If the time is wrong, it is set once and tested again.
    """

    def get_time():
        result = command.run_check("hwclock --utc --show")[0].strip()
        return datetime.strptime(result, "%Y-%m-%d %H:%M:%S.%f+0:00")

    def set_time(time):
        time = time.strftime("%Y-%m-%d %H:%M:%S.%f+0:00")
        command.run_check(f'hwclock --utc --set --date "{time}"')

    offset = abs((get_time() - datetime.utcnow()).total_seconds())
    if offset > 60:
        set_time(datetime.utcnow())
        offset = abs((get_time() - datetime.utcnow()).total_seconds())
    assert offset < 60



from labgrid.remote.scheduler import *


def test_simple():
    places = [
        TagSet("place-1", {"name=place-1", "soc=mx6"}),
        TagSet("place-2", {"name=place-2", "soc=mx8"}),
    ]
    filters = [
        TagSet("res-1", {"soc=mx8"}),
        TagSet("res-2", {"soc=mx6"}),
    ]
    expected = {
        "res-1": "place-2",
        "res-2": "place-1",
    }

    assert schedule(places, filters) == expected
    assert schedule(places, filters[::-1]) == expected
    assert schedule(places[::-1], filters) == expected
    assert schedule(places[::-1], filters[::-1]) == expected


def test_overlap():
    places = [
        TagSet("place-1", {"name=place-1", "soc=mx6", "arch=arm"}),
        TagSet("place-2", {"name=place-2", "soc=mx8", "arch=arm"}),
    ]
    filters = [
        TagSet("res-1", {"arch=arm"}),
        TagSet("res-2", {"soc=mx6"}),
    ]
    expected = {
        "res-1": "place-2",
        "res-2": "place-1",
    }

    assert schedule(places, filters) == expected
    assert schedule(places, filters[::-1]) == expected
    assert schedule(places[::-1], filters) == expected
    assert schedule(places[::-1], filters[::-1]) == expected


def test_flexible():
    places = [
        TagSet("place-1", {"name=place-1", "board=foo"}),
        TagSet("place-2", {"name=place-2", "board=foo"}),
    ]
    filters = [
        TagSet("res-1", {"board=foo"}),
        TagSet("res-2", {"board=foo"}),
    ]
    expected_keys = {"res-1", "res-2"}
    expected_vals = {"place-1", "place-2"}

    allocation = schedule(places, filters)
    assert allocation.keys() == expected_keys
    assert set(allocation.values()) == expected_vals
    allocation = schedule(places, filters[::-1])
    assert allocation.keys() == expected_keys
    assert set(allocation.values()) == expected_vals
    allocation = schedule(places[::-1], filters)
    assert allocation.keys() == expected_keys
    assert set(allocation.values()) == expected_vals
    allocation = schedule(places[::-1], filters[::-1])
    assert allocation.keys() == expected_keys
    assert set(allocation.values()) == expected_vals


def test_conflict():
    places = [
        TagSet("place-1", {"name=place-1", "board=foo"}),
    ]
    filters = [
        TagSet("res-1", {"board=foo"}),
        TagSet("res-2", {"board=foo"}),
    ]

    assert schedule(places, filters) == {"res-1": "place-1"}
    assert schedule(places, filters[::-1]) == {"res-2": "place-1"}
    assert schedule(places[::-1], filters) == {"res-1": "place-1"}
    assert schedule(places[::-1], filters[::-1]) == {"res-2": "place-1"}



"""
Test labgrid code of communicating with a docker daemon and using it
for creating, starting and accessing a docker container.
"""

import pytest
import io

from labgrid import Environment
from labgrid.driver import DockerDriver
from labgrid.resource.docker import DockerConstants
from labgrid.exceptions import NoResourceFoundError

pytest.importorskip("docker")


def check_external_progs_present():
    """Determine if host machine has a usable docker daemon"""
    try:
        import docker

        try:
            dock = docker.from_env()
            dock.info()
        except docker.errors.DockerException:
            return False
    except OSError:
        return False
    return True


@pytest.fixture
def docker_env(tmp_path_factory):
    """Create Environment instance from the given inline YAML file."""
    p = tmp_path_factory.mktemp("docker") / "config.yaml"
    # Note: The SSHDriver part at bottom is only used by the test that
    # will run if a docker daemon is present, not by
    # test_driver_without_daemon.
    p.write_text(
        """
        targets:
          main:
            resources:
            - DockerDaemon:
                docker_daemon_url: "unix:///var/run/docker.sock"
            drivers:
            - DockerDriver:
                image_uri: "rastasheep/ubuntu-sshd:16.04"
                pull: 'missing'
                container_name: "ubuntu-lg-example"
                host_config: {"network_mode": "bridge"}
                network_services: [
                  {"port": 22, "username": "root", "password": "root"}]
            - DockerStrategy: {}
            - SSHDriver:
                keyfile: ""
        """
    )
    return Environment(str(p))


@pytest.fixture
def docker_env_for_local_container(tmp_path_factory):
    """Create Environment instance from the given inline YAML file."""
    p = tmp_path_factory.mktemp("docker") / "config.yaml"
    p.write_text(
        """
        targets:
          main:
            resources:
            - DockerDaemon:
                docker_daemon_url: "unix:///var/run/docker.sock"
            drivers:
            - DockerDriver:
                image_uri: "local_rastasheep"
                pull: "never"
                container_name: "ubuntu-lg-example"
                host_config: {"network_mode": "bridge"}
                network_services: [
                  {"port": 22, "username": "root", "password": "root"}]
            - DockerStrategy: {}
            - SSHDriver:
                keyfile: ""
        """
    )
    return Environment(str(p))


@pytest.fixture
def docker_target(docker_env):
    """Get a labgrid Target instance from the Environment instance
    given by docker_env. When tearing down the Target instance, make sure
    singleton ResourceManager is "reset".
    """
    t = docker_env.get_target()
    yield t

    # Fake! In real life, ResourceManager is intended to be a singleton.
    # The class is created only once - when python parses common.py.
    # But this means that the class with its "instances" attribute survives
    # from test case to test case. This is not what we want. On the contrary,
    # we want each of test_docker_with_daemon and test_docker_without_daemon
    # to run with a *fresh* instance of the ResourceManager singleton.
    #
    # Luckily it is easy to "reset" ResourceManager. The singleton is kept
    # in attribute "instances" so by resetting "instances" to {}, next test
    # case will force creation of a fresh ResourceManager instance.
    from labgrid.resource import ResourceManager

    ResourceManager.instances = {}


@pytest.fixture
def command(docker_target):
    """Bring system to a state where it's possible to execute commands
    on a running docker container. When done, stop the container again.
    """
    strategy = docker_target.get_driver("DockerStrategy")
    strategy.transition("accessible")
    shell = docker_target.get_driver("CommandProtocol")
    yield shell
    strategy.transition("gone")


@pytest.fixture
def docker_target_for_local_image(docker_env_for_local_container):
    """Same as `docker_target` but uses a different image uri"""
    t = docker_env_for_local_container.get_target()
    yield t

    from labgrid.resource import ResourceManager

    ResourceManager.instances = {}


@pytest.fixture
def local_command(docker_target_for_local_image):
    """Same as `command` but uses a different image uri"""
    strategy = docker_target_for_local_image.get_driver("DockerStrategy")
    strategy.transition("accessible")
    shell = docker_target_for_local_image.get_driver("CommandProtocol")
    yield shell
    strategy.transition("gone")


@pytest.mark.skipif(not check_external_progs_present(), reason="No access to a docker daemon")
def test_docker_with_daemon(command):
    """Test the docker machinery when a running docker daemon can be used
    (thus test is skipped if there is no such daemon). The tests executes
    a few tests inside a running docker container using SSHDriver for access.
    """
    stdout, stderr, return_code = command.run("cat /proc/version")
    assert return_code == 0
    assert len(stdout) > 0
    assert len(stderr) == 0
    assert "Linux" in stdout[0]

    stdout, stderr, return_code = command.run("false")
    assert return_code != 0
    assert len(stdout) == 0
    assert len(stderr) == 0


@pytest.fixture
def build_image():
    import docker

    client = docker.from_env()
    dockerfile_content = """
    FROM rastasheep/ubuntu-sshd:16.04
    """
    dockerfile_stream = io.BytesIO(dockerfile_content.encode("utf-8"))
    image, logs = client.images.build(fileobj=dockerfile_stream, tag="local_rastasheep", rm=True)


@pytest.mark.skipif(not check_external_progs_present(), reason="No access to a docker daemon")
def test_docker_with_daemon_and_local_image(build_image, local_command):
    """Build a container locally and connect to it"""
    stdout, stderr, return_code = local_command.run("cat /proc/version")
    assert return_code == 0
    assert len(stdout) > 0
    assert len(stderr) == 0
    assert "Linux" in stdout[0]

    stdout, stderr, return_code = local_command.run("false")
    assert return_code != 0
    assert len(stdout) == 0
    assert len(stderr) == 0


def test_create_driver_fail_missing_docker_daemon(target):
    """The test target does not contain any DockerDaemon instance -
    and so creation must fail.
    """
    with pytest.raises(NoResourceFoundError):
        DockerDriver(target, "docker_driver")


def test_docker_without_daemon(docker_env, mocker):
    """Test as many aspects as possible of DockerDriver, DockerDaemon,
    DockerManager and DockerStrategy without using an actual
    docker daemon, real sockets or system time"""

    import docker

    # Target::update_resources() and Target::await_resources use
    # time.monotonic() and time.sleep() to control when to search
    # for resources. Avoid time delays and make running from cmd-line
    # and inside debugger equal by mocking out all time.
    time_monotonic = mocker.patch("labgrid.target.monotonic")
    time_monotonic.side_effect = [1.0, 2.0, 3.0, 4.0, 5.0, 6.0]

    # Mock actions on the imported "docker" python module
    docker_client_class = mocker.patch("docker.DockerClient", autospec=True)
    docker_client = docker_client_class.return_value
    api_client_class = mocker.patch("docker.api.client.APIClient", autospec=True)
    docker_client.api = api_client_class.return_value
    api_client = docker_client.api
    api_client.base_url = "unix:///var/run/docker.sock"
    # First, a "mocked" old docker container is returned by
    # ...api.containers(); this is done when DockerDaemon tries
    # to clean up old containers. Next, a one-item list is delivered by
    # ...api.containers() which is part of DockerDaemon::update_resources()
    # - it is cached for future use; therefore no need to replicate
    # this entry in the side_effects list.
    api_client.containers.side_effect = [
        [
            {
                "Labels": {DockerConstants.DOCKER_LG_CLEANUP_LABEL: DockerConstants.DOCKER_LG_CLEANUP_TYPE_AUTO},
                "NetworkSettings": {"IPAddress": "1.1.1.1"},
                "Names": "old-one",
                "Id": "0",
            }
        ],
        [
            {
                "Labels": {DockerConstants.DOCKER_LG_CLEANUP_LABEL: DockerConstants.DOCKER_LG_CLEANUP_TYPE_AUTO},
                "NetworkSettings": {"IPAddress": "2.1.1.1"},
                "Names": "actual-one",
                "Id": "1",
            }
        ],
    ]
    docker_client.images.get.side_effect = docker.errors.ImageNotFound(
        "Image not found", response=None, explanation=""
    )

    # Mock actions on the imported "socket" python module
    socket_create_connection = mocker.patch("socket.create_connection")
    sock = mocker.MagicMock()
    # First two negative connection setup attempts are used at initial
    # resource setup during strategy.transition("shell"); these simulate
    # that it takes time for the docker container to come up. The final,
    # successful, return value is delivered when t.update_resources()
    # is called explicitly later on.
    socket_create_connection.side_effect = [
        Exception("No connection on first call"),
        Exception("No connection on second call"),
        sock,
    ]

    # get_target() - which calls make_target() - creates resources/drivers
    # from .yaml configured environment. Creation provokes binding
    # and attempts to connect to network services.
    api_client.remove_container.assert_not_called()
    t = docker_env.get_target()
    assert api_client.remove_container.call_count == 1

    # Make sure DockerDriver didn't accidentally succeed with a socket connect
    # attempt (this fact is actually expressed by what happens next -
    # the socket is closed).
    sock.shutdown.assert_not_called()
    sock.close.assert_not_called()

    # Get strategy - needed to transition to "shell" state.
    strategy = t.get_driver("DockerStrategy")

    # strategy starts in state "unknown" so the following should be a no-op.
    strategy.transition("unknown")

    # Now activate DockerDriver and set it "on". This creates and starts
    # a (mocked) container.
    strategy.transition("accessible")

    # Assert what mock calls transitioning to "shell" must have caused
    #
    # DockerDriver::on_activate():
    image_uri = t.get_driver("DockerDriver").image_uri
    docker_client.images.get.assert_called_once_with(image_uri)
    docker_client.images.pull.assert_called_once_with(image_uri)

    assert api_client.create_host_config.call_count == 1
    assert api_client.create_container.call_count == 1
    #
    # DockerDriver::on()
    assert api_client.start.call_count == 1

    # From here the test using the real docker daemon would proceed with
    #   shell = t.get_driver('CommandProtocol')
    #   shell.run('...')
    # which makes use of e.g. the SSHDriver.  Binding the SSHDriver
    # is important since it triggers activation of the NetworkService.
    # But then SSHDriver uses ssh to connect to the NetworkService
    # which will lead to error. Instead just call update_resources()
    # directly - which is what is needed to provoke DockerDaemon to create
    # a new NetworkService instance.
    t.update_resources()

    # This time socket connection was successful
    # (per the third socket_create_connection return value defined above).
    assert sock.shutdown.call_count == 1
    assert sock.close.call_count == 1

    # Bonus: Test what happens if taking a forbidden strategy transition:
    # "shell" -> "unknown".
    from labgrid.strategy import StrategyError

    with pytest.raises(StrategyError):
        strategy.transition("unknown")

    # test and reset broken state
    assert isinstance(strategy.broken, StrategyError)
    strategy.broken = None

    # Also bonus: How are invalid state names handled?
    with pytest.raises(StrategyError):
        strategy.transition("this is not a valid state")

    # test and reset broken state
    assert isinstance(strategy.broken, KeyError)
    strategy.broken = None

    # Return to "gone" state - to also use that part of the DockerDriver code.
    strategy.transition("gone")
    from labgrid.strategy.dockerstrategy import Status

    assert strategy.status == Status.gone



import attr

from .common import Driver
from ..factory import target_factory
from ..step import step
from .exception import ExecutionError
from ..util.helper import processwrapper


@target_factory.reg_driver
@attr.s(eq=False)
class LXAUSBMuxDriver(Driver):
    """The LXAUSBMuxDriver uses the usbmuxctl tool to control the USBMux
    hardware
    """

    bindings = {
        "mux": {"LXAUSBMux", "NetworkLXAUSBMux"},
    }

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        if self.target.env:
            self.tool = self.target.env.config.get_tool("usbmuxctl")
        else:
            self.tool = "usbmuxctl"

    @Driver.check_active
    @step(title="usbmux_set", args=["links"])
    def set_links(self, links):
        args = []
        for link in links:
            link = link.lower()
            if link == "dut-device":
                args.append("--dut-device")
            elif link == "host-dut":
                args.append("--host-dut")
            elif link == "host-device":
                args.append("--host-device")
            else:
                raise ExecutionError(f"Link '{link}' not supported by LXAUSBMuxDriver")

        cmd = self.mux.command_prefix + [
            self.tool,
            "--path",
            self.mux.path,
            "connect",
            *args,
        ]
        processwrapper.check_output(cmd)



import pytest


@pytest.fixture(scope="session")
def command(target):
    strategy = target.get_driver("DockerStrategy")
    strategy.transition("accessible")
    shell = target.get_driver("CommandProtocol")
    return shell



def test_modbusrtu_example(instrument):
    """
    The interface for the modbus RTU driver is a thin adapter, having same
    interface as provided by the minimalmodbus package. Therefore, for more
    infromation on how to use the labgrid modbus RTU driver, please refer to
    the minimalmodbus documentation.
    """

    uptime_register_addr = 0x107
    instrument.write_register(uptime_register_addr, 0, functioncode=6)
    value = instrument.read_register(uptime_register_addr)

    assert value <= 1



import attr

from ..factory import target_factory
from .common import Resource


@attr.s(eq=False)
class BaseProvider(Resource):
    internal = attr.ib(validator=attr.validators.instance_of(str))
    external = attr.ib(validator=attr.validators.instance_of(str))

    def __attrs_post_init__(self):
        self.host = "localhost"
        super().__attrs_post_init__()


@target_factory.reg_resource
@attr.s(eq=False)
class TFTPProvider(BaseProvider):
    pass


@target_factory.reg_resource
@attr.s(eq=False)
class NFSProvider(Resource):
    def __attrs_post_init__(self):
        self.host = "localhost"
        super().__attrs_post_init__()


@target_factory.reg_resource
@attr.s(eq=False)
class HTTPProvider(BaseProvider):
    pass



"""
This module implements the communication protocol to switch the digital outputs
on a "dcttech" USB Relay.

Supported modules:

- USBRelay2 (and likely others)

Supported Functionality:

- Turn digital output on and off
"""

import errno
from contextlib import contextmanager
from time import monotonic, sleep

import usb.core
import usb.util

GET_REPORT = 0x1
SET_REPORT = 0x9
REPORT_TYPE_FEATURE = 3


class USBHIDRelay:
    def __init__(self, **args):
        self._dev = usb.core.find(**args)
        if self._dev is None:
            raise ValueError("Device not found")

        if self._dev.idVendor == 0x16C0:
            self._set_output = self._set_output_dcttech
            self._get_output = self._get_output_dcttech
        elif self._dev.idVendor == 0x5131:
            self._set_output = self._set_output_lcus
            self._get_output = self._get_output_lcus
        else:
            raise ValueError(f"Unknown vendor/protocol for VID {self._dev.idVendor:x}")

        if self._dev.is_kernel_driver_active(0):
            self._dev.detach_kernel_driver(0)

    @contextmanager
    def _claimed(self):
        timeout = monotonic() + 1.0
        while True:
            try:
                usb.util.claim_interface(self._dev, 0)
                break
            except usb.core.USBError as e:
                if monotonic() > timeout:
                    raise e
                if e.errno == errno.EBUSY:
                    sleep(0.01)
                else:
                    raise e
        yield
        usb.util.release_interface(self._dev, 0)

    def _set_output_dcttech(self, number, status):
        assert 1 <= number <= 8
        req = [0xFF if status else 0xFD, number]
        self._dev.ctrl_transfer(
            usb.util.CTRL_TYPE_CLASS | usb.util.CTRL_RECIPIENT_DEVICE | usb.util.ENDPOINT_OUT,
            SET_REPORT,
            (REPORT_TYPE_FEATURE << 8) | 0,  # no report ID
            0,
            req,  # payload
        )

    def _get_output_dcttech(self, number):
        assert 1 <= number <= 8
        resp = self._dev.ctrl_transfer(
            usb.util.CTRL_TYPE_CLASS | usb.util.CTRL_RECIPIENT_DEVICE | usb.util.ENDPOINT_IN,
            GET_REPORT,
            (REPORT_TYPE_FEATURE << 8) | 0,  # no report ID
            0,
            8,  # size
        )
        return bool(resp[7] & (1 << (number - 1)))

    def _set_output_lcus(self, number, status):
        assert 1 <= number <= 8
        ep_in = self._dev[0][(0, 0)][0]
        ep_out = self._dev[0][(0, 0)][1]
        req = [0xA0, number, 0x01 if status else 0x00, 0x00]
        req[3] = sum(req) & 0xFF
        ep_out.write(req)
        ep_in.read(64)

    def _get_output_lcus(self, number):
        assert 1 <= number <= 8
        # we have no information on how to read the current value
        return False

    def set_output(self, number, status):
        with self._claimed():
            self._set_output(number, status)

    def get_output(self, number):
        with self._claimed():
            return self._get_output(number)


_relays = {}


def _get_relay(busnum, devnum):
    if (busnum, devnum) not in _relays:
        _relays[(busnum, devnum)] = USBHIDRelay(bus=busnum, address=devnum)
    return _relays[(busnum, devnum)]


def handle_set(busnum, devnum, number, status):
    relay = _get_relay(busnum, devnum)
    relay.set_output(number, status)


def handle_get(busnum, devnum, number):
    relay = _get_relay(busnum, devnum)
    return relay.get_output(number)


methods = {
    "set": handle_set,
    "get": handle_get,
}



import pytest


def test_bootchooser(command):
    stdout, stderr, returncode = command.run("bootchooser -i")
    if returncode == 127:
        pytest.skip("bootchooser command not available")
    assert returncode == 0
    assert not stderr
    assert stdout[0].startswith("Good targets")
    assert stdout[1] != "none"



import shlex

import attr
from pexpect import TIMEOUT

from ..factory import target_factory
from ..protocol import CommandProtocol, ConsoleProtocol, LinuxBootProtocol
from ..step import step
from ..util import gen_marker, Timeout, re_vt100
from .common import Driver
from .commandmixin import CommandMixin


@target_factory.reg_driver
@attr.s(eq=False)
class BareboxDriver(CommandMixin, Driver, CommandProtocol, LinuxBootProtocol):
    """BareboxDriver - Driver to control barebox via the console.
       BareboxDriver binds on top of a ConsoleProtocol.

       On activation, the BareboxDriver will look for the barebox prompt on the
       console, stopping any autoboot counters if necessary, and provide access
       to the barebox shell.

    Args:
        prompt (str): barebox prompt to match
        autoboot (regex): optional, autoboot message to match
        interrupt (str): optional, string to interrupt autoboot (use "\x03" for CTRL-C)
        bootstring (regex): optional, regex indicating that the Linux Kernel is booting
        password (str): optional, password to use for access to the shell
        boot_command (str): optional, boot command to boot target
        login_timeout (int): optional, timeout for access to the shell
    """
    bindings = {"console": ConsoleProtocol, }
    prompt = attr.ib(default="", validator=attr.validators.instance_of(str))
    autoboot = attr.ib(default="stop autoboot", validator=attr.validators.instance_of(str))
    interrupt = attr.ib(default="\x04", validator=attr.validators.instance_of(str))
    bootstring = attr.ib(default=r"Linux version \d", validator=attr.validators.instance_of(str))
    password = attr.ib(default="", validator=attr.validators.instance_of(str))
    boot_command = attr.ib(default="boot -v", validator=attr.validators.instance_of(str))
    login_timeout = attr.ib(default=60, validator=attr.validators.instance_of(int))

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self._status = 0
        # barebox' default log level, used as fallback if no log level can be saved
        self.saved_log_level = 7

    def on_activate(self):
        """Activate the BareboxDriver

        This function tries to login if not already active
        """
        if self._status == 0:
            self._await_prompt()

    def on_deactivate(self):
        """Deactivate the BareboxDriver

        Simply sets the internal status to 0
        """
        self._status = 0

    @Driver.check_active
    @step(args=['cmd'])
    def run(self, cmd: str, *, timeout: int = 30):
        return self._run(cmd, timeout=timeout)

    def _run(self, cmd: str, *, timeout: int = 30, adjust_log_level: bool = True, codec: str = "utf-8", decodeerrors: str = "strict"):  # pylint: disable=unused-argument,line-too-long
        """
        Runs the specified command on the shell and returns the output.

        Args:
            cmd (str): command to run on the shell
            timeout (int): optional, timeout in seconds

        Returns:
            Tuple[List[str],List[str], int]: if successful, None otherwise
        """
        # FIXME: use codec, decodeerrors
        marker = gen_marker()
        # hide marker from expect
        hidden_marker = f'"{marker[:4]}""{marker[4:]}"'
        # generate command with marker and log level adjustment
        cmp_command = f'echo -o /cmd {shlex.quote(cmd)}; echo {hidden_marker};'
        if self.saved_log_level and adjust_log_level:
            cmp_command += f' global.loglevel={self.saved_log_level};'
        cmp_command += f' sh /cmd; echo {hidden_marker} $?;'
        if self.saved_log_level and adjust_log_level:
            cmp_command += ' global.loglevel=0;'

        if self._status == 1:
            self.console.sendline(cmp_command)
            _, _, match, _ = self.console.expect(
                rf'{marker}(.*){marker}\s+(\d+)\s+.*{self.prompt}',
                timeout=timeout)
            # Remove VT100 Codes and split by newline
            data = re_vt100.sub('', match.group(1).decode('utf-8')).split('\r\n')[1:-1]
            self.logger.debug("Received Data: %s", data)
            # Get exit code
            exitcode = int(match.group(2))
            return (data, [], exitcode)

        return None

    @Driver.check_active
    @step()
    def reset(self):
        """Reset the board via a CPU reset
        """
        self._status = 0
        self.console.sendline("reset")
        self._await_prompt()

    def get_status(self):
        """Retrieve status of the BareboxDriver
        0 means inactive, 1 means active.

        Returns:
            int: status of the driver
        """
        return self._status

    def _check_prompt(self):
        """
        Internal function to check if we have a valid prompt.
        It sets the internal _status to 1 or 0 based on the prompt detection.
        """
        marker = gen_marker()
        # hide marker from expect
        hidden_marker = f'"{marker[:4]}""{marker[4:]}"'
        self.console.sendline(f"echo {hidden_marker}")
        try:
            self.console.expect(f"{marker}", timeout=2)
            self.console.expect(self.prompt, timeout=1)
            self._status = 1
        except TIMEOUT:
            self._status = 0
            raise

    @step()
    def _await_prompt(self):
        """Awaits the prompt and enters the shell"""

        timeout = Timeout(float(self.login_timeout))

        # We call console.expect with a short timeout here to detect if the
        # console is idle, which would result in a timeout without any changes
        # to the before property. So we store the last before value we've seen.
        # Because pexpect keeps any read data in it's buffer when a timeout
        # occours, we can't lose any data this way.
        last_before = None
        password_entered = False

        expectations = [self.prompt, self.autoboot, "Password: ", TIMEOUT]
        while True:
            index, before, _, _ = self.console.expect(
                expectations,
                timeout=2
            )

            if index == 0:
                # we got a prompt. no need for any further action to activate
                # this driver.
                self._status = 1
                break

            elif index == 1:
                # we need to interrupt autoboot
                self.console.write(self.interrupt.encode('ASCII'))

            elif index == 2:
                # we need to enter the password
                if not self.password:
                    raise Exception("Password entry needed but no password set")
                if password_entered:
                    # we already sent the password, but got the pw prompt again
                    raise Exception("Password was not correct")
                self.console.sendline(self.password)
                password_entered = True

            elif index == 3:
                # expect hit a timeout while waiting for a match
                if before == last_before:
                    # we did not receive anything during the previous expect cycle
                    # let's assume the target is idle and we can safely issue a
                    # newline to check the state
                    self.console.sendline("")

                if timeout.expired:
                    raise TIMEOUT(
                        f"Timeout of {self.login_timeout} seconds exceeded during waiting for login"
                    )

            last_before = before

        self._check_prompt()

        # remember barebox' log level - we don't expect to be interrupted here
        # by pollers because no hardware interaction is triggered by echo, so
        # it should be safe to use the usual shell wrapper via _run()
        stdout, _, exitcode = self._run("echo $global.loglevel", adjust_log_level=False)
        [saved_log_level] = stdout
        if exitcode == 0 and saved_log_level.isnumeric():
            self.saved_log_level = saved_log_level

        # silence barebox, the driver can get confused by asynchronous messages
        # logged to the console otherwise
        self._run("global.loglevel=0", adjust_log_level=False)

    @Driver.check_active
    def await_boot(self):
        """Wait for the initial Linux version string to verify we successfully
        jumped into the kernel.
        """
        self.console.expect(self.bootstring)

    @Driver.check_active
    def boot(self, name: str = ""):
        """Boot the default or a specific boot entry

        Args:
            name (str): name of the entry to boot"""
        # recover saved log level
        self._run(f"global.loglevel={self.saved_log_level}", adjust_log_level=False)

        if name:
            self.console.sendline(f"boot -v {name}")
        else:
            self.console.sendline(self.boot_command)



from urllib.parse import urlparse

import pytest

from labgrid.driver import ExecutionError
from labgrid.driver.power.poe_netgear_plus import _get_hostname_and_password
from labgrid.resource import NetworkPowerPort, YKUSHPowerPort
from labgrid.driver.powerdriver import (
    ExternalPowerDriver,
    ManualPowerDriver,
    NetworkPowerDriver,
    YKUSHPowerDriver,
)


class TestManualPowerDriver:
    def test_create(self, target):
        d = ManualPowerDriver(target, 'foo-board')
        assert (isinstance(d, ManualPowerDriver))

    def test_on(self, target, mocker):
        m = mocker.patch('builtins.input')

        d = ManualPowerDriver(target, 'foo-board')
        target.activate(d)
        d.on()

        m.assert_called_once_with(
            "Turn the target Test ON and press enter"
        )

    def test_off(self, target, mocker):
        m = mocker.patch('builtins.input')

        d = ManualPowerDriver(target, 'foo-board')
        target.activate(d)
        d.off()

        m.assert_called_once_with(
            "Turn the target Test OFF and press enter"
        )

    def test_cycle(self, target, mocker):
        m = mocker.patch('builtins.input')

        d = ManualPowerDriver(target, 'foo-board')
        target.activate(d)
        d.cycle()

        m.assert_called_once_with("CYCLE the target Test and press enter")


class TestExternalPowerDriver:
    def test_create(self, target):
        d = ExternalPowerDriver(
            target, 'power', cmd_on='set -1 foo-board', cmd_off='set -0 foo-board'
        )
        assert (isinstance(d, ExternalPowerDriver))

    def test_on(self, target, mocker):
        pty = mocker.patch('pty.openpty')
        mocker.patch('fcntl.fcntl')
        fdopen = mocker.patch('os.fdopen')
        close = mocker.patch('os.close')
        popen = mocker.patch('subprocess.Popen')
        select = mocker.patch('select.select')
        fd_mock = mocker.MagicMock()
        instance_mock = mocker.MagicMock()
        select_mock = mocker.MagicMock()
        popen.return_value = instance_mock
        fdopen.return_value = fd_mock
        pty.return_value = (instance_mock, 2)
        fd_mock.read.return_value = b'Done\nDone'
        instance_mock.returncode = 0
        select.return_value = ([], [], [])

        d = ExternalPowerDriver(
            target, 'power', cmd_on='set -1 foo-board', cmd_off='set -0 foo-board'
        )
        target.activate(d)
        d.on()

        popen.assert_called_once_with(['set', '-1', 'foo-board'], bufsize=0,
                                      stderr=2, stdout=2)

    def test_off(self, target, mocker):
        pty = mocker.patch('pty.openpty')
        mocker.patch('fcntl.fcntl')
        fdopen = mocker.patch('os.fdopen')
        close = mocker.patch('os.close')
        popen = mocker.patch('subprocess.Popen')
        select = mocker.patch('select.select')
        fd_mock = mocker.MagicMock()
        instance_mock = mocker.MagicMock()
        select_mock = mocker.MagicMock()
        popen.return_value = instance_mock
        fdopen.return_value = fd_mock
        pty.return_value = (instance_mock, 2)
        fd_mock.read.return_value = b'Done\nDone'
        instance_mock.returncode = 0
        select.return_value = ([], [], [])

        d = ExternalPowerDriver(
            target, 'power', cmd_on='set -1 foo-board', cmd_off='set -0 foo-board'
        )
        target.activate(d)
        d.off()

        popen.assert_called_once_with(['set', '-0', 'foo-board'], bufsize=0,
                                      stderr=2, stdout=2)

    def test_cycle(self, target, mocker):
        pty = mocker.patch('pty.openpty')
        mocker.patch('fcntl.fcntl')
        fdopen = mocker.patch('os.fdopen')
        close = mocker.patch('os.close')
        popen = mocker.patch('subprocess.Popen')
        select = mocker.patch('select.select')
        fd_mock = mocker.MagicMock()
        instance_mock = mocker.MagicMock()
        select_mock = mocker.MagicMock()
        popen.return_value = instance_mock
        fdopen.return_value = fd_mock
        pty.return_value = (instance_mock, 2)
        fd_mock.read.return_value = b'Done\nDone'
        instance_mock.returncode = 0
        select.return_value = ([], [], [])

        m_sleep = mocker.patch('time.sleep')

        d = ExternalPowerDriver(
            target, 'power', cmd_on='set -1 foo-board', cmd_off='set -0 foo-board'
        )
        target.activate(d)
        d.cycle()

        assert popen.call_args_list == [
            mocker.call(['set', '-0', 'foo-board'], bufsize=0, stderr=2,
                        stdout=2),
            mocker.call(['set', '-1', 'foo-board'], bufsize=0, stderr=2,
                        stdout=2),
        ]
        m_sleep.assert_called_once_with(2.0)

    def test_cycle_explicit(self, target, mocker):
        pty = mocker.patch('pty.openpty')
        mocker.patch('fcntl.fcntl')
        fdopen = mocker.patch('os.fdopen')
        close = mocker.patch('os.close')
        popen = mocker.patch('subprocess.Popen')
        select = mocker.patch('select.select')
        fd_mock = mocker.MagicMock()
        instance_mock = mocker.MagicMock()
        select_mock = mocker.MagicMock()
        popen.return_value = instance_mock
        fdopen.return_value = fd_mock
        pty.return_value = (instance_mock, 2)
        fd_mock.read.return_value = b'Done\nDone'
        instance_mock.returncode = 0
        select.return_value = ([], [], [])

        m_sleep = mocker.patch('time.sleep')

        d = ExternalPowerDriver(
            target, 'power',
            cmd_on='set -1 foo-board',
            cmd_off='set -0 foo-board',
            cmd_cycle='set -c foo-board',
        )
        target.activate(d)
        d.cycle()

        popen.assert_called_once_with(['set', '-c', 'foo-board'], bufsize=0,
                                      stderr=2, stdout=2)
        m_sleep.assert_not_called()

class TestNetworkPowerDriver:
    def test_create(self, target):
        r = NetworkPowerPort(target, 'power', model='netio', host='dummy', index='1')
        d = NetworkPowerDriver(target, 'power')
        assert isinstance(d, NetworkPowerDriver)

    @pytest.mark.parametrize('backend', ('rest', 'simplerest'))
    @pytest.mark.parametrize(
        'host',
        (
            'http://example.com/{index}',
            'https://example.com/{index}',
            'http://example.com:1234/{index}',
            'https://example.com:1234/{index}',
            'http://user:pass@example.com:1234/{index}',
            'https://user:pass@example.com:1234/{index}',
        )
    )
    def test_create_backend_with_url_in_host(self, target, mocker, backend, host):
        get = mocker.patch('requests.get')
        get.return_value.text = '1'
        mocker.patch('requests.put')

        index = '1'
        NetworkPowerPort(target, 'power', model=backend, host=host, index=index)
        d = NetworkPowerDriver(target, 'power')
        assert isinstance(d, NetworkPowerDriver)
        target.activate(d)

        d.cycle()
        assert d.get() is True

        # the called URL should be similar to the one configured in the resource, but with
        # index and explicit port
        expected_host = host.format(index=index)
        url = urlparse(expected_host)
        if url.port is None:
            implicit_port = 443 if url.scheme == 'https' else 80
            expected_host = expected_host.replace(url.netloc, f'{url.netloc}:{implicit_port}')

        get.assert_called_with(expected_host)

    @pytest.mark.parametrize(
        'host',
        (
            'http://example.com',
            'https://example.com',
        )
    )
    def test_create_shelly_gen1_backend_with_url_in_host(self, target, mocker, host):
        get = mocker.patch('requests.get')
        get.return_value.text = '{"ison": true}'
        mocker.patch('requests.post')

        index = '0'
        NetworkPowerPort(target, 'power', model='shelly_gen1', host=host, index=index)
        d = NetworkPowerDriver(target, 'power')
        target.activate(d)

        d.cycle()
        assert d.get() is True

        # the called URL should be similar to the one configured in the resource, but with
        # index and explicit port
        expected_host = f"{host}/relay/{index}"
        url = urlparse(expected_host)
        if url.port is None:
            implicit_port = 443 if url.scheme == "https" else 80
            expected_host = expected_host.replace(
                url.netloc, f"{url.netloc}:{implicit_port}"
            )

        get.assert_called_with(expected_host)

    def test_create_ubus_backend(self, target, mocker):
        post = mocker.patch("requests.post")
        post.return_value.json.return_value = {
            "jsonrpc": "2.0",
            "id": 1,
            "result": [
                0,
                {
                    "firmware": "v80.1",
                    "budget": 77.000000,
                    "consumption": 1.700000,
                    "ports": {
                        "lan1": {
                            "priority": 0,
                            "mode": "PoE+",
                            "status": "Delivering power",
                            "consumption": 1.700000,
                        }
                    },
                },
            ],
        }

        index = "1"
        NetworkPowerPort(
            target, "power", model="ubus", host="http://example.com/ubus", index=index
        )
        d = NetworkPowerDriver(target, "power")
        target.activate(d)

        d.cycle()
        assert d.get() is True

    def test_import_backends(self):
        import labgrid.driver.power
        import labgrid.driver.power.apc
        import labgrid.driver.power.digipower
        import labgrid.driver.power.digitalloggers_http
        import labgrid.driver.power.digitalloggers_restapi
        import labgrid.driver.power.eth008
        import labgrid.driver.power.gude
        import labgrid.driver.power.gude24
        import labgrid.driver.power.netio
        import labgrid.driver.power.netio_kshell
        import labgrid.driver.power.pe6216
        import labgrid.driver.power.poe_netgear_plus
        import labgrid.driver.power.rest
        import labgrid.driver.power.sentry
        import labgrid.driver.power.eg_pms2_network
        import labgrid.driver.power.shelly_gen1
        import labgrid.driver.power.ubus
        import labgrid.driver.power.tinycontrol_tcpdu

    def test_import_backend_eaton(self):
        pytest.importorskip("pysnmp")
        import labgrid.driver.power.eaton

    def test_import_backend_tplink(self):
        pytest.importorskip("kasa")
        import labgrid.driver.power.tplink

    def test_import_backend_siglent(self):
        pytest.importorskip("vxi11")
        import labgrid.driver.power.siglent

    def test_import_backend_poe_mib(self):
        pytest.importorskip("pysnmp")
        import labgrid.driver.power.poe_mib

class TestYKUSHPowerDriver:
    YKUSH_FAKE_SERIAL = "YK12345"
    YKUSH_LIST_OUTPUT = f"Attached YKUSH Boards:\n1. Board found with serial number: {YKUSH_FAKE_SERIAL}".encode(
        "utf-8"
    )
    YKUSH3_FAKE_SERIAL = "Y3N10673"
    YKUSH3_LIST_OUTPUT = f"Attached YKUSH3 Boards:\n1. Board found with serial number: {YKUSH3_FAKE_SERIAL}".encode(
        "utf-8"
    )
    YKUSHXS_LIST_OUTPUT = (
        "Attached YKUSH XS Boards:\n1. Board found with serial number: YKU1234".encode(
            "utf-8"
        )
    )

    def test_create(self, target):
        resource = YKUSHPowerPort(
            target, "power", serial=self.YKUSH_FAKE_SERIAL, index=1
        )
        device = YKUSHPowerDriver(target, "power")
        assert isinstance(device, YKUSHPowerDriver)

    def test_default_off(self, target, mocker):
        check_output_mock = mocker.patch(
            "labgrid.util.helper.processwrapper.check_output"
        )
        check_output_mock.side_effect = [
            self.YKUSH_LIST_OUTPUT,
            self.YKUSHXS_LIST_OUTPUT,
            self.YKUSH3_LIST_OUTPUT,
            b"",
        ]
        resource = YKUSHPowerPort(
            target, "power", serial=self.YKUSH_FAKE_SERIAL, index=2
        )
        resource.avail = True
        device = YKUSHPowerDriver(target, "power")
        target.activate(device)
        device.off()

        check_output_mock.assert_called_with(
            ["ykushcmd", "ykush", "-s", self.YKUSH_FAKE_SERIAL, "-d", "2"]
        )

    def test_ykush3_on(self, target, mocker):
        check_output_mock = mocker.patch(
            "labgrid.util.helper.processwrapper.check_output"
        )
        check_output_mock.side_effect = [
            self.YKUSH_LIST_OUTPUT,
            self.YKUSHXS_LIST_OUTPUT,
            self.YKUSH3_LIST_OUTPUT,
            b"",
        ]
        resource = YKUSHPowerPort(
            target, "power", serial=self.YKUSH3_FAKE_SERIAL, index=3
        )
        resource.avail = True
        device = YKUSHPowerDriver(target, "power")
        target.activate(device)
        device.on()

        check_output_mock.assert_called_with(
            ["ykushcmd", "ykush3", "-s", self.YKUSH3_FAKE_SERIAL, "-u", "3"]
        )


class TestPoeNetgearPlusPowerDriver:
    @pytest.mark.parametrize(
        'url, expected_host, expected_pw',
        [
            ("http://example.com", "example.com", "P4ssword"),
            ("http://ignored:detected_pw@example.com", "example.com", "detected_pw"),
        ]
    )
    def test_get_hostname_and_password(self, url: str, expected_host: str, expected_pw: str):
        returned_host, returned_pw = _get_hostname_and_password(url)
        assert returned_host == expected_host
        assert returned_pw == expected_pw

    def test_get_hostname_and_pw_non_http_raises(self):
        with pytest.raises(ExecutionError, match="URL must start with http://"):
            _get_hostname_and_password("no_http_protocol")



from importlib import import_module

import attr

from ..factory import target_factory
from ..protocol import DigitalOutputProtocol
from ..util.proxy import proxymanager
from .common import Driver
from .exception import ExecutionError
from ..step import step


@target_factory.reg_driver
@attr.s(eq=False)
class LXAIOBusPIODriver(Driver, DigitalOutputProtocol):
    bindings = {
        "pio": {"LXAIOBusPIO", "NetworkLXAIOBusPIO"},
    }

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self._requests = import_module("requests")

    def on_activate(self):
        # we can only forward if the backend knows which port to use
        host, port = proxymanager.get_host_and_port(self.pio)
        self._url = f"http://{host}:{port}/nodes/{self.pio.node}/pins/{self.pio.pin}/"

    @Driver.check_active
    @step(args=["status"])
    def set(self, status):
        if self.pio.invert:
            status = not status
        r = self._requests.post(self._url, data={"value": "1" if status else "0"})
        r.raise_for_status()
        j = r.json()
        if j["code"] != 0:
            raise ExecutionError(f"failed to set value: {j['error_message']}")

    @Driver.check_active
    @step(result=["True"])
    def get(self):
        r = self._requests.get(self._url)
        r.raise_for_status()
        j = r.json()
        if j["code"] != 0:
            raise ExecutionError(f"failed to get value: {j['error_message']}")
        result = j["result"]
        if result not in (0, 1):
            raise ExecutionError(f"invalid input value: {repr(result)}")
        status = bool(result)
        if self.pio.invert:
            status = not status
        return status



import attr

from ..factory import target_factory
from .common import Resource


@target_factory.reg_resource
@attr.s(eq=False)
class AndroidNetFastboot(Resource):
    address = attr.ib(validator=attr.validators.instance_of(str))
    port = attr.ib(default=5554, validator=attr.validators.instance_of(int))
    protocol = attr.ib(default="udp", validator=attr.validators.instance_of(str))



import pytest

from labgrid.driver import QEMUDriver
from labgrid import Environment

@pytest.fixture
def qemu_env(tmpdir):
    p = tmpdir.join("config.yaml")
    p.write(
        """
        targets:
          main:
            role: foo
        images:
          kernel: "test.zImage"
          dtb: test.dtb"
        tools:
          qemu: "qemu-system-arm"
        paths:
          rootfs: "test/path"
        """
    )
    return Environment(str(p))

@pytest.fixture
def qemu_target(qemu_env):
    return qemu_env.get_target()

@pytest.fixture
def qemu_driver(qemu_target):
    q = QEMUDriver(
        qemu_target,
        "qemu",
        qemu_bin="qemu",
        machine='',
        cpu='',
        memory='',
        boot_args='',
        extra_args='',
        kernel='kernel',
        rootfs='rootfs')
    return q

@pytest.fixture
def qemu_mock(mocker):
    popen_mock = mocker.patch('subprocess.Popen')
    popen_mock.return_value.wait.return_value = 0
    popen_mock.return_value.stdout.readline.return_value = b"""
    {
      "QMP": {
        "version": {}
      },
      "return": {}
    }
    """

    select_mock = mocker.patch('select.select')
    select_mock.return_value = True, None, None

    socket_mock = mocker.patch('socket.socket')
    socket_mock.return_value.accept.return_value = mocker.MagicMock(), ''

@pytest.fixture
def qemu_version_mock(mocker):
    run_mock = mocker.patch('subprocess.run')
    run_mock.return_value.returncode = 0
    run_mock.return_value.stdout = "QEMU emulator version 4.2.1"

def test_qemu_instance(qemu_target, qemu_driver):
    assert (isinstance(qemu_driver, QEMUDriver))

def test_qemu_activate_deactivate(qemu_target, qemu_driver, qemu_version_mock):
    qemu_target.activate(qemu_driver)
    qemu_target.deactivate(qemu_driver)

def test_qemu_on_off(qemu_target, qemu_driver, qemu_mock, qemu_version_mock):
    qemu_target.activate(qemu_driver)

    qemu_driver.on()
    qemu_driver.off()

    qemu_target.deactivate(qemu_driver)

def test_qemu_read_write(qemu_target, qemu_driver, qemu_mock, qemu_version_mock):
    qemu_target.activate(qemu_driver)

    qemu_driver.on()
    qemu_driver.read()
    qemu_driver.read(max_size=10)
    qemu_driver.write(b'abc')

    qemu_target.deactivate(qemu_driver)

def test_qemu_port_forwarding(qemu_target, qemu_driver, qemu_mock, qemu_version_mock):
    qemu_target.activate(qemu_driver)

    qemu_driver.on()
    qemu_driver.add_port_forward('tcp', '127.0.0.1', 8080, '127.0.0.1', 80)
    assert ('tcp', '127.0.0.1', 8080, '') in qemu_driver._forwarded_ports.keys()
    qemu_driver.remove_port_forward('tcp', '127.0.0.1', 8080)
    assert qemu_driver._forwarded_ports == {}

    qemu_target.deactivate(qemu_driver)

def test_qemu_port_forwarding_with_netdev(qemu_target, qemu_driver, qemu_mock, qemu_version_mock):
    qemu_target.activate(qemu_driver)

    qemu_driver.on()
    qemu_driver.add_port_forward('tcp', '127.0.0.1', 8080, '127.0.0.1', 80, netdev='netdev0')
    assert ('tcp', '127.0.0.1', 8080, 'netdev0') in qemu_driver._forwarded_ports.keys()
    qemu_driver.remove_port_forward('tcp', '127.0.0.1', 8080, netdev='netdev0')
    assert qemu_driver._forwarded_ports == {}

    qemu_target.deactivate(qemu_driver)



import attr

from ..factory import target_factory
from .common import Resource


@target_factory.reg_resource
@attr.s(eq=False)
class ModbusTCPCoil(Resource):
    """This resource describes Modbus TCP coil.

    Args:
        host (str): hostname of the Modbus TCP server e.g. "192.168.23.42:502"
        coil (int): index of the coil e.g. 3
        invert (bool): optional, whether the logic level is be inverted (active-low)
        write_multiple_coils (bool): optional, whether write using multiple coils method"""

    host = attr.ib(validator=attr.validators.instance_of(str))
    coil = attr.ib(validator=attr.validators.instance_of(int))
    invert = attr.ib(default=False, validator=attr.validators.instance_of(bool))
    write_multiple_coils = attr.ib(default=False, validator=attr.validators.instance_of(bool))


@target_factory.reg_resource
@attr.s(eq=False)
class WaveshareModbusTCPCoil(ModbusTCPCoil):
    """This resource describes Waveshare brand Modbus TCP coil.

    Args:
        host (str): hostname of the Modbus TCP server e.g. "192.168.23.42:502"
        coil (int): index of the coil e.g. 3
        coil_count (int): The total number of coils on this module.
        invert (bool): optional, whether the logic level is be inverted (active-low)
        write_multiple_coils (bool): optional, whether write using multiple coils method"""

    coil_count = attr.ib(default=8, validator=attr.validators.instance_of(int))



from time import monotonic

from pytest import approx


def test_sleep(command):
    # measure the round-trip-time
    timestamp = monotonic()
    stdout, stderr, returncode = command.run("true")
    elapsed_true = monotonic() - timestamp
    assert returncode == 0
    assert not stdout
    assert not stderr

    timestamp = monotonic()
    stdout, stderr, returncode = command.run("sleep 1")
    elapsed_sleep = monotonic() - timestamp
    assert returncode == 0
    assert not stdout
    assert not stderr

    assert elapsed_true < elapsed_sleep

    assert elapsed_sleep - elapsed_true == approx(1.0, abs=1e-2)
